from textual.app import ComposeResult, Widget
from textual.containers import Horizontal
from textual.widgets import Label
from textual.color import Color

from ai_hydra.constants.DHydraTui import DField
from ai_hydra.utils.HydraMetrics import HydraMetrics
from ai_hydra.utils.MetricEvent import MemEvent

BUCKET = "bucket"
NUM_BUCKETS = 20
BASE_COLOR = "#0f1e10"


def percent_to_hex(p: float) -> str:
    """
    Convert a float in [0.0, 1.0] to a hex string "00" - "ff".
    """
    p = max(0.0, min(1.0, p))
    value = int(round(p * 255))
    return f"{value:02x}"


class ATHMemory(Widget):

    def __init__(self, metrics: HydraMetrics, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.metrics = metrics

    def compose(self) -> ComposeResult:
        with Horizontal():
            yield Label("Memory", id=DField.REPLAY_MEM)
            for i in range(NUM_BUCKETS):
                yield Label(id=f"b{i}", classes=BUCKET)

    def refresh_data(self) -> None:

        mem_event = self.metrics.get_bucket_snaphot()
        if mem_event is None:
            return

        bucket_counts = mem_event.bucket_counts
        max_size = max(bucket_counts)
        min_size = min(bucket_counts)
        span = max_size - min_size

        bucket_idx = 0
        for count in bucket_counts:
            label_id = f"b{bucket_idx}"
            cur_label = self.query_one(f"#{label_id}", Label)
            cur_label.update(str(count))

            pct = 1.0 if span == 0 else (count - min_size) / span
            alpha_hex = percent_to_hex(pct)
            cur_label.styles.background = Color.parse(
                f"{BASE_COLOR}{alpha_hex}"
            )
            bucket_idx += 1



# ai_hydra/constants/DEpsilonNice.py
#
#    AI Hydra
#    Author: Nadim-Daniel Ghaznavi
#    Copyright: (c) 2025-2026 Nadim-Daniel Ghaznavi
#    GitHub: https://github.com/NadimGhaznavi/ai_hydra
#    Website: https://ai-hydra.readthedocs.io/en/latest
#    License: GPL 3.0

from typing import Final

from ai_hydra.constants.DHydraTui import DField


class DEpsilonNice:
    WINDOW: Final[str] = "window"
    EPOCH: Final[str] = "epoch"
    CALLS: Final[str] = "calls"
    TRIGGERED: Final[str] = "triggered"
    FATAL_SUGGESTED: Final[str] = "fatal_suggested"
    OVERRIDES: Final[str] = "overrides"
    NO_SAFE_ALTERNATIVE: Final[str] = "no_safe_alternative"
    TRIGGER_RATE: Final[str] = "trigger_rate"
    OVERRIDE_RATE: Final[str] = "override_rate"



# ai_hydra/nnet/RecurrentTrainer.py
#
#    AI Hydra
#    Author: Nadim-Daniel Ghaznavi
#    Copyright: (c) 2025-2026 Nadim-Daniel Ghaznavi
#    GitHub: https://github.com/NadimGhaznavi/ai_hydra
#    Website: https://ai-hydra.readthedocs.io/en/latest
#    License: GPL 3.0


import numpy as np
import torch
import torch.nn as nn
import torch.optim as optim
from copy import deepcopy

from ai_hydra.constants.DHydraTui import DField
from ai_hydra.constants.DHydra import DHydraLog, DModule
from ai_hydra.constants.DNNet import DRecurrentTrainer

from ai_hydra.nnet.ATH.ATHMemory import ATHMemory
from ai_hydra.utils.HydraLog import HydraLog

GRAD_CLIPPING = True
DQN_WITH_TARGET = False
DOUBLE_DQN = True


class RecurrentTrainer:
    def __init__(
        self,
        model,
        replay: ATHMemory,
        lr: float,
        log_level: DHydraLog,
        device: torch.device | None = None,
    ):
        self.device = device or torch.device(DField.CPU)
        self.model = model.to(self.device)

        if DQN_WITH_TARGET or DOUBLE_DQN:
            self.target_model = deepcopy(model).to(self.device)
            self.target_model.eval()

        self.replay = replay
        self._tau = None
        self._gamma = None

        self.optimizer = DRecurrentTrainer.OPTIM(
            self.model.parameters(), lr=lr
        )
        self.criterion = DRecurrentTrainer.CRITERION()

        self.log = HydraLog(
            client_id=DModule.RECURRENT_TRAINER,
            log_level=log_level,
            to_console=True,
        )

        self._losses: list[float] = []
        self._cold_memory = True
        if DQN_WITH_TARGET and DOUBLE_DQN:
            raise ValueError(
                "You cannot enable both DQN_WITH_TARGET and DOUBLE_DQN"
            )
        self.log.info(f"Setting the learning rate to {lr}")
        self.log.debug("Initialized")

    def get_avg_loss(self) -> float | None:
        if not self._losses:
            return None
        avg_loss = sum(self._losses) / len(self._losses)
        self._losses = []
        return avg_loss

    def reset(self):
        self.optimizer.state = {}
        if DQN_WITH_TARGET or DOUBLE_DQN:
            self.target_model = deepcopy(self.model).to(self.device)
            self.target_model.eval()

    async def train_long_memory(self) -> float | None:

        chunks = await self.replay.data_mgr.sample_chunks()
        if chunks is None:
            return

        if self._cold_memory:
            self._cold_memory = False
            self.log.debug(
                f"Training with {len(chunks)} batches with sequence length {len(chunks[0])}"
            )

        # Shapes (B = batch_size, T = seq_length, F = feature/input size)
        # states      -> [B, T, F]
        # actions     -> [B, T]
        # rewards     -> [B, T]
        # next_states -> [B, T, F]
        # dones       -> [B, T]
        states = torch.tensor(
            np.array(
                [[t.old_state for t in chunk] for chunk in chunks],
                dtype=np.float32,
            ),
            dtype=torch.float32,
            device=self.device,
        )
        actions = torch.tensor(
            np.array(
                [[t.action for t in chunk] for chunk in chunks], dtype=np.int64
            ),
            dtype=torch.long,
            device=self.device,
        )
        rewards = torch.tensor(
            np.array(
                [[t.reward for t in chunk] for chunk in chunks],
                dtype=np.float32,
            ),
            dtype=torch.float32,
            device=self.device,
        )
        next_states = torch.tensor(
            np.array(
                [[t.new_state for t in chunk] for chunk in chunks],
                dtype=np.float32,
            ),
            dtype=torch.float32,
            device=self.device,
        )
        dones = torch.tensor(
            np.array(
                [[t.done for t in chunk] for chunk in chunks], dtype=np.float32
            ),
            dtype=torch.float32,
            device=self.device,
        )

        self.model.train()

        q_pred_all = self.model.forward_sequence(states)  # [B, T, A]
        q_pred = q_pred_all.gather(2, actions.unsqueeze(-1)).squeeze(
            -1
        )  # [B, T]

        with torch.no_grad():
            if DOUBLE_DQN:
                self.model.eval()  # Don't include dropout noise
                # [B, T, A]
                online_next_q = self.model.forward_sequence(next_states)
                self.model.train()
                next_actions = online_next_q.argmax(
                    dim=2, keepdim=True
                )  # [B, T, 1]

                # target net evaluates those actions
                target_next_q = self.target_model.forward_sequence(
                    next_states
                )  # [B, T, A]
                max_next_q = target_next_q.gather(2, next_actions).squeeze(
                    -1
                )  # [B, T]

            elif DQN_WITH_TARGET:
                # [B, T, A]
                next_q = self.target_model.forward_sequence(next_states)
                max_next_q = next_q.max(dim=2).values  # [B, T]

            else:
                # Don't include dropout noise (eval()/train() toggle)
                self.model.eval()
                next_q = self.model.forward_sequence(next_states)
                self.model.train()
                max_next_q = next_q.max(dim=2).values  # [B, T]

            q_target = rewards + self._gamma * max_next_q * (1.0 - dones)

        loss = self.criterion(q_pred, q_target)

        self.optimizer.zero_grad()
        loss.backward()

        if GRAD_CLIPPING:
            torch.nn.utils.clip_grad_norm_(
                self.model.parameters(), max_norm=1.0
            )

        self.optimizer.step()

        if DQN_WITH_TARGET or DOUBLE_DQN:
            self._soft_update_target()

        self._losses.append(loss.item())
        return float(loss.item())

    def _soft_update_target(self) -> None:
        for target_param, param in zip(
            self.target_model.parameters(),
            self.model.parameters(),
        ):
            target_param.data.copy_(
                self._tau * param.data + (1.0 - self._tau) * target_param.data
            )

    def set_params(self, tau: float, gamma: float):
        self._tau = tau
        self.log.info(f"Setting Tau to {tau}")
        self._gamma = gamma
        self.log.info(f"Setting the Discount/Gamma to {gamma}")



# ai_hydra/constants/DHydraMQ.py
#
#    AI Hydra
#    Author: Nadim-Daniel Ghaznavi
#    Copyright: (c) 2025-2026 Nadim-Daniel Ghaznavi
#    GitHub: https://github.com/NadimGhaznavi/ai_hydra
#    Website: https://ai-hydra.readthedocs.io/en/latest
#    License: GPL 3.0

from typing import Final


class DEvent:
    DATA: Final[str] = "data"
    EV_TYPE: Final[str] = "ev_type"
    LEVEL: Final[str] = "level"
    MESSAGE: Final[str] = "message"
    PAYLOAD: Final[str] = "payload"
    SENDER: Final[str] = "sender"


class DHydraMQ:
    HEARTBEAT: Final[str] = "heartbeat"
    MSGS: Final[str] = "msgs"
    PER_EP: Final[str] = "per_ep"
    PER_STEP: Final[str] = "per_step"
    SCORES: Final[str] = "scores"
    TIMER: Final[str] = "timer"
    UTF_8: Final[str] = "utf-8"


class DHydraMQDef:
    TOPIC_PREFIX: Final[str] = "ai_hydra"
    PER_STEP_TOPIC: Final[str] = "per_step_topic"
    PER_EPISODE_TOPIC: Final[str] = "per_episode_topic"
    EVENTS_TOPIC: Final[str] = "events_topic"
    SCORES_TOPIC: Final[str] = "scores_topic"
    MAX_BATCH_TIME: Final[float] = 0.5
    MAX_BATCH_SIZE: Final[int] = 100



# ai_hydra/game/GameLogic.py
#
#    AI Hydra
#    Author: Nadim-Daniel Ghaznavi
#    Copyright: (c) 2025-2026 Nadim-Daniel Ghaznavi
#    GitHub: https://github.com/NadimGhaznavi/ai_hydra
#    Website: https://ai-hydra.readthedocs.io/en/latest
#    License: GPL 3.0
#

"""
Game logic for the Pong game.
"""

from __future__ import annotations

import random
from typing import List, Tuple

from ai_hydra.constants.DGame import DGameDef, DGameField
from ai_hydra.game.GameBoard import (
    Direction,
    GameBoard,
    Move,
    MoveAction,
    Position,
)
from ai_hydra.game.MoveResult import MoveResult


class PongLogic:

    @staticmethod
    def step(
        board: GameBoard,
        action: MoveAction | int,
        rng: random.Random,
        *,
        food_ends_episode: bool = False,
    ) -> MoveResult:
        """
        Execute one action and return a MoveResult containing the new board.

        Notes:
          - rng is the ONLY source of randomness.
          - board is treated as immutable; a new GameBoard is returned.
        """

        # Normalize action
        if isinstance(action, str):
            action = int(action)

        if isinstance(action, int):
            if action == 0:
                action = MoveAction.LEFT_TURN
            elif action == 1:
                action = MoveAction.STRAIGHT
            elif action == 2:
                action = MoveAction.RIGHT_TURN
            else:
                raise ValueError(f"Invalid action index: {action}")

        # Build the move (fixes the undefined 'move' bug)
        move = PongLogic.create_move(board.direction, action)

        # Compute new head position
        new_head = Position(
            board.snake_head.x + move.resulting_direction.dx,
            board.snake_head.y + move.resulting_direction.dy,
        )



# ai_hydra/nnet/EpsilonAlgo.py
#
#    AI Hydra
#    Author: Nadim-Daniel Ghaznavi
#    Copyright: (c) 2025-2026 Nadim-Daniel Ghaznavi
#    GitHub: https://github.com/NadimGhaznavi/ai_hydra
#    Website: https://ai-hydra.readthedocs.io/en/latest
#    License: GPL 3.0

from __future__ import annotations

import zmq.asyncio

import random
from typing import Optional, Callable, Awaitable

from ai_hydra.constants.DNNet import DEpsilonField
from ai_hydra.constants.DHydra import DHydraLog, DModule
from ai_hydra.constants.DEvent import EV_STATUS

from ai_hydra.utils.HydraLog import HydraLog
from ai_hydra.zmq.HydraEventMQ import EventMsg, HydraEventMQ


class EpsilonAlgo:
    """
    Epsilon-greedy helper.

    - Call maybe_random_action() each step.
      If it returns an int action (0..2), take it.
      If it returns None, use your policy's greedy action.

    - Call played_game() at end of episode to decay epsilon.
    """

    def __init__(
        self,
        rng: random.Random,
        log_level: DHydraLog,
        pub_func,
    ):
        self._rng = rng

        self.event = HydraEventMQ(
            client_id=DModule.EPSILON_ALGO, pub_func=pub_func
        )

        self._initial_epsilon = None
        self._min_epsilon = None
        self._decay_rate = None

        self._cur_epsilon = None
        self._injected = 0
        self._depleted = False
        self._epsilon_not_depleted = True
        self.log = HydraLog(
            client_id=DModule.EPSILON_ALGO,
            log_level=log_level,
            to_console=True,
        )

    def cur_epsilon(self) -> float:
        return self._cur_epsilon

    def decay_rate(self, value: float) -> None:
        self._decay_rate = value
        self.log.debug(f"Epsilon decay rate set: {value}")

    def get_params(self) -> dict[str, float]:
        return {
            DEpsilonField.INITIAL: self._initial_epsilon,
            DEpsilonField.MINIMUM: self._min_epsilon,
            DEpsilonField.DECAY_RATE: self._decay_rate,
        }

    def initial_epsilon(self, value: float) -> None:
        self._initial_epsilon = value
        self._cur_epsilon = value
        self.log.debug(f"Initial epsilon set: {value}")

    def injected(self) -> int:
        return self._injected

    def maybe_random_action(self) -> Optional[int]:
        """
        Returns:
            int action in {0,1,2} if we inject randomness, else None.
        """
        if self._rng.random() < self._cur_epsilon:
            self._injected += 1
            return self._rng.randrange(3)
        return None

    def min_epsilon(self, value: float) -> None:
        self._min_epsilon = value
        self.log.debug(f"Minimum epsilon set: {value}")

    async def played_game(self) -> None:
        """
        Decay epsilon at the end of an episode.
        """
        self._cur_epsilon = max(
            self._min_epsilon, self._cur_epsilon * self._decay_rate
        )
        self._depleted = self._cur_epsilon <= self._min_epsilon

        if self._cur_epsilon < 0.001:
            self._cur_epsilon = 0.0

        if self._epsilon_not_depleted and self._depleted:
            msg = f"Exploration complete: ε = {self._min_epsilon}"
            self.log.info(msg)
            await self.event.publish(
                EventMsg(level=EV_STATUS.INFO, message=msg)
            )
            self._epsilon_not_depleted = False

        self.reset_injected()

    def reset(self) -> None:
        """
        Start a fresh schedule (called at START_RUN or RESET)
        """
        self._cur_epsilon = self._initial_epsilon
        self._depleted = self._cur_epsilon <= self._min_epsilon
        self.reset_injected()

    def reset_injected(self) -> None:
        """
        Reset the "injected random moves" count.
        """
        self._injected = 0

    def set_params(
        self, *, initial: float, minimum: float, decay: float
    ) -> None:
        self._initial_epsilon = float(initial)
        self._min_epsilon = float(minimum)
        self._decay_rate = float(decay)



# ai_hydra/client/plots/ScoreDistPlot.py
#
#    AI Hydra
#    Author: Nadim-Daniel Ghaznavi
#    Copyright: (c) 2025-2026 Nadim-Daniel Ghaznavi
#    GitHub: https://github.com/NadimGhaznavi/ai_hydra
#    Website: https://ai-hydra.readthedocs.io/en/latest
#    License: GPL 3.0

from statistics import mean, median

from textual.app import ComposeResult, Widget
from textual.containers import Horizontal
from textual_plot import HiResMode, LegendLocation, PlotWidget

from ai_hydra.constants.DHydraTui import DField, DLabel, DColor, DPlotDef
from ai_hydra.utils.HydraMetrics import HydraMetrics


class ScoresDistPlot(Widget):

    def __init__(self, metrics: HydraMetrics, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.metrics = metrics

    def compose(self) -> ComposeResult:
        yield Horizontal(
            PlotWidget(id=DField.PLOT_SCORES_DIST),
            PlotWidget(id=DField.PLOT_RECENT_SCORES_DIST),
        )

    def clear(self):
        self.query_one(f"#{DField.PLOT_SCORES_DIST}").clear()
        self.query_one(f"#{DField.PLOT_RECENT_SCORES_DIST}").clear()
        self.plot_all()

    def plot_all(self):
        self.plot_scores_dist()
        self.plot_recent_scores_dist()

    def plot_scores_dist(self):
        scores, counts = self.metrics.get_scores_dist_plot_points()
        stats = self.metrics.get_scores_dist_stats()
        self._plot(tag=DField.ALL, scores=scores, counts=counts, stats=stats)

    def plot_recent_scores_dist(self):
        scores, counts = self.metrics.get_recent_scores_dist_plot_points()
        stats = self.metrics.get_recent_scores_dist_stats()
        self._plot(
            tag=DField.RECENT, scores=scores, counts=counts, stats=stats
        )

    def _plot(self, tag, scores, counts, stats):
        if not scores:
            return

        if tag == DField.ALL:
            plot = self.query_one(f"#{DField.PLOT_SCORES_DIST}", PlotWidget)
            ylabel = DLabel.COUNT
        elif tag == DField.RECENT:
            plot = self.query_one(
                f"#{DField.PLOT_RECENT_SCORES_DIST}", PlotWidget
            )
            ylabel = f"Recent ({DPlotDef.RECENT_SCORES_MAX}) Count"

        plot.clear()
        plot.bar(
            scores,
            counts,
            bar_style=DColor.BLUE,
            width=0.8,
            hires_mode=HiResMode.BRAILLE,
        )

        plot.set_ylimits()
        plot.set_xlabel(DLabel.SCORES)
        plot.set_ylabel(ylabel)
        plot.show_legend(location=LegendLocation.TOPRIGHT)

        if stats is None:
            return

        mean_score, median_score = stats

        plot.add_v_line(
            mean_score, DColor.GREEN, f"{DLabel.MEAN}  : {mean_score:.2f}"
        )
        plot.add_v_line(
            median_score, DColor.PURPLE, f"{DLabel.MEDIAN}: {median_score:.2f}"
        )

        cur_epoch = self.metrics.get_cur_epoch()

        if cur_epoch > 0 and cur_epoch % 500 == 0:
            if tag == DField.ALL:
                self.metrics.add_mean_median(
                    epoch=cur_epoch, mean=mean_score, median=median_score
                )
            elif tag == DField.RECENT:
                self.metrics.add_recent_mean_and_median(
                    epoch=cur_epoch, mean=mean_score, median=median_score
                )



# ai_hydra/client/TabbedPlots.py
#
#    AI Hydra
#    Author: Nadim-Daniel Ghaznavi
#    Copyright: (c) 2025-2026 Nadim-Daniel Ghaznavi
#    GitHub: https://github.com/NadimGhaznavi/ai_hydra
#    Website: https://ai-hydra.readthedocs.io/en/latest
#    License: GPL 3.0

from __future__ import annotations

from collections import deque
from statistics import median, mean
from typing import Iterable

from textual import on
from textual.app import ComposeResult, Widget
from textual.widgets import TabbedContent
from textual.containers import Horizontal
from textual_plot import HiResMode, LegendLocation, PlotWidget

from ai_hydra.constants.DHydraTui import DColor, DField, DLabel, DPlotDef

from ai_hydra.client.plots.PlotUtils import (
    rolling_mean,
    thin_series,
    weighted_mean,
    weighted_median,
)


class TabbedPlots(Widget):
    """Tabbed plot widget for game scores, loss, and scatter visualizations."""

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._reset_data()

    def compose(self) -> ComposeResult:
        with TabbedContent(
            DLabel.GAME_SCORES,
            DLabel.SCORES,
            DLabel.LOSS,
        ):
            yield PlotWidget(id=DField.GAME_SCORES)
            yield PlotWidget(id=DField.SCORES_PLOT)
            yield Horizontal(
                PlotWidget(id=DField.LOSS_PLOT), PlotWidget(id=DField.CUR_LOSS)
            )

    def action_show_tab(self, tab: str) -> None:
        self.get_child_by_type(TabbedContent).active = tab

    @on(TabbedContent.TabActivated)
    def handle_tab_activated(self, event: TabbedContent.TabActivated) -> None:
        pane_id = event.pane.id
        self.call_later(lambda: self._replot_active_tab(pane_id))

    def add_loss(self, epoch: int, loss: float, plot: bool = True) -> None:
        self.loss_epochs.append(epoch)
        self.losses.append(loss)

        self.cur_losses.append(loss)
        self.cur_epochs.append(epoch)

        if plot:
            self._plot_loss()

    def add_score(
        self, cur_score: int | None, cur_epoch: int | None, plot: bool = True
    ) -> None:
        if cur_score is None or cur_epoch is None:
            return

        score = int(cur_score)
        epoch = int(cur_epoch)

        self.score_counts[score] = self.score_counts.get(score, 0) + 1
        self.game_scores.append(score)
        self.game_epochs.append(epoch)

        if plot:
            self._plot_scores()
            self._plot_game_scores()

    def add_scatter_score(
        self,
        scores: tuple[int, int] | list[tuple[int, int]],
        plot: bool = True,
    ) -> None:
        """Add one or more scatter points as (episode, score)."""
        pass

    def reset(self) -> None:
        self._reset_data()

        self.query_one(f"#{DField.GAME_SCORES}", PlotWidget).clear()
        self.query_one(f"#{DField.SCORES_PLOT}", PlotWidget).clear()
        self.query_one(f"#{DField.LOSS_PLOT}", PlotWidget).clear()
        self.query_one(f"#{DField.CUR_LOSS}", PlotWidget).clear()

    def _reset_data(self) -> None:
        max_points = DPlotDef.MAX_CUR_DATA_POINTS
        max_loss_points = DPlotDef.MAX_CUR_LOSS_DATA_POINTS

        self.loss_epochs: list[int] = []
        self.losses: list[float] = []

        self.score_counts: dict[int, int] = {}

        self.game_scores: deque[int] = deque(maxlen=max_points)
        self.game_epochs: deque[int] = deque(maxlen=max_points)
        self.avg_game_score: list[float] = []

        self.cur_losses: deque[float] = deque(maxlen=max_loss_points)
        self.cur_epochs: deque[int] = deque(maxlen=max_loss_points)

    def _replot_active_tab(self, pane_id: str) -> None:
        if pane_id == "tab-1":
            self._plot_game_scores()
        elif pane_id == "tab-2":
            self._plot_scores()
        elif pane_id == "tab-3":
            self._plot_loss()
        else:
            raise ValueError(f"Unhandled tab: {pane_id}")

    def _plot_game_scores(self) -> None:
        plot = self.query_one(f"#{DField.GAME_SCORES}", PlotWidget)
        plot.clear()

        if not self.game_epochs or not self.game_scores:
            return

        epochs = list(self.game_epochs)
        scores = list(self.game_scores)

        # Thin only if the series gets too dense for the viewport.
        plot_epochs = epochs
        plot_scores = scores
        if len(scores) > DPlotDef.MAX_CUR_DATA_POINTS:
            plot_epochs, plot_scores = thin_series(
                x=epochs,
                y=scores,
                max_points=DPlotDef.MAX_CUR_DATA_POINTS,
            )

        # Current score as a line plot.
        plot.plot(
            x=plot_epochs,
            y=plot_scores,
            hires_mode=HiResMode.BRAILLE,
            line_style=DColor.GREEN,
            label=DLabel.CURRENT,
        )

        # Rolling average / trend line.
        window = max(8, len(scores) // DPlotDef.AVG_DIVISOR)
        if len(scores) >= window:
            smoothed_scores = rolling_mean(scores, window)
            smoothed_epochs = epochs[window - 1 :]

            plot.plot(
                x=smoothed_epochs,
                y=smoothed_scores,
                hires_mode=HiResMode.BRAILLE,
                line_style=DColor.RED,
                label=DLabel.AVERAGE,
            )

        plot.set_xlabel(DLabel.EPISODES)
        plot.set_ylabel(DLabel.SCORES)
        plot.set_ylimits()
        plot.show_legend(location=LegendLocation.TOPLEFT)

    def _plot_scores(self) -> None:
        plot = self.query_one(f"#{DField.SCORES_PLOT}", PlotWidget)
        plot.clear()

        if not self.score_counts:
            return

        x_values = sorted(self.score_counts.keys())
        y_values = [self.score_counts[score] for score in x_values]

        plot.bar(
            x_values,
            y_values,
            bar_style=DColor.GREEN,
            width=0.8,
            hires_mode=HiResMode.BRAILLE,
        )

        plot.set_ylimits()
        plot.set_xlabel(DLabel.SCORES)
        plot.set_ylabel(DLabel.COUNT)
        plot.show_legend(location=LegendLocation.TOPRIGHT)

        mean_score = mean(self.score_counts)
        median_score = median(self.score_counts)

        plot.add_v_line(mean_score, DColor.GREEN, f"Mean  : {mean_score:.2f}")
        plot.add_v_line(median_score, "purple", f"Median: {median_score:.2f}")

    def _plot_loss(self) -> None:
        self._plot_full_loss()
        self._plot_current_loss()

    def _plot_full_loss(self) -> None:
        plot = self.query_one(f"#{DField.LOSS_PLOT}", PlotWidget)
        plot.clear()

        if not self.loss_epochs or not self.losses:
            return

        epochs = self.loss_epochs
        losses = self.losses

        if len(losses) > DPlotDef.MAX_LOSS_DATA_POINTS:
            epochs, losses = thin_series(
                x=epochs,
                y=losses,
                max_points=DPlotDef.MAX_LOSS_DATA_POINTS,
            )

        plot.plot(
            x=epochs,
            y=losses,
            hires_mode=HiResMode.BRAILLE,
            line_style=DColor.GREEN,
            label=DLabel.COMPLETE,
        )
        plot.set_ylimits()
        plot.set_xlabel(DLabel.EPISODES)
        plot.set_ylabel(DLabel.LOSS)
        plot.show_legend(location=LegendLocation.TOPLEFT)

        # Add vertical lines to mark high score events
        # for event in self.scatter_scores:
        #    episode, score = event
        #    plot.add_v_line(episode, DColor.GREEN)

    def _plot_current_loss(self) -> None:
        plot = self.query_one(f"#{DField.CUR_LOSS}", PlotWidget)
        plot.clear()

        if not self.cur_epochs or not self.cur_losses:
            return

        epochs = list(self.cur_epochs)
        losses = list(self.cur_losses)

        plot.plot(
            x=epochs,
            y=losses,
            hires_mode=HiResMode.BRAILLE,
            line_style=DColor.GREEN,
            label=DLabel.CURRENT,
        )
        plot.set_ylimits()
        plot.set_xlabel(DLabel.EPISODES)
        plot.set_ylabel(DLabel.LOSS)
        plot.show_legend(location=LegendLocation.BOTTOMLEFT)



# ai_hydra/constants/DReplayMemory.py
#
#    AI Hydra
#    Author: Nadim-Daniel Ghaznavi
#    Copyright: (c) 2025-2026 Nadim-Daniel Ghaznavi
#    GitHub: https://github.com/NadimGhaznavi/ai_hydra
#    Website: https://ai-hydra.readthedocs.io/en/latest
#    License: GPL 3.0

from typing import Final


class DMemField:
    MAX_BUCKETS: Final[str] = "max_buckets"
    MAX_FRAMES: Final[str] = "max_frames"
    THRESHOLDS_REQUIRED: Final[str] = "thresholds_required"


class DMemDef:
    # Used by the ATHGearbox to calculate the seq_length/batch_size for the
    # different gears.
    MAX_TRAINING_FRAMES: Final[int] = 512
    # The highest gear
    MAX_GEAR: Final[int] = 26
    # ATH memory stores this maximum number of frames
    MAX_FRAMES: Final[int] = 125000
    # ATH uses this many memory buckets
    MAX_BUCKETS: Final[int] = 20
    # The ATHGearbox won't shift up or down before NUM_COOLDOWN_EPISODES
    # episodes have gone by.
    NUM_COOLDOWN_EPISODES: Final[int] = 100
    # If the THRESHOLD_BUCKETS have a combined contents greater than
    # UPSHIFT_COUNT_THRESHOLD, the the ATHGearbox shifts UP.
    # If the THRESHOLD_BUCKETS have a combined contents less than
    # DOWNSHIFT_COUNT_THRESHOLD, the the ATHGearbox shifts DOWN.
    THRESHOLD_BUCKETS: Final[list] = [17, 18, 19]
    UPSHIFT_COUNT_THRESHOLD: Final[int] = 150
    DOWNSHIFT_COUNT_THRESHOLD: Final[int] = 50
    # The number of episodes that the TrainMgr will tolerate with no new
    # highscore before calling `stagnation_warning()` on the ATHReplay.
    MAX_STAGNANT_EPISODES: Final[int] = 300
    # The number of episodes without a new highscore before the TrainMgr
    # calls `hard_reset()`.
    MAX_HARD_RESET_EPISODES: Final[int] = 2000


class DMemory:
    # Simple Replay Memory - Linear model
    MIN_FRAMES: Final[int] = 1500
    MAX_MEM_SIZE: Final[int] = 75000



# ai_hydra/client/plots/GameScorePlot.py
#
#    AI Hydra
#    Author: Nadim-Daniel Ghaznavi
#    Copyright: (c) 2025-2026 Nadim-Daniel Ghaznavi
#    GitHub: https://github.com/NadimGhaznavi/ai_hydra
#    Website: https://ai-hydra.readthedocs.io/en/latest
#    License: GPL 3.0

from textual.app import ComposeResult, Widget
from textual.containers import Horizontal
from textual_plot import HiResMode, LegendLocation, PlotWidget

from ai_hydra.constants.DHydraTui import DField, DLabel, DColor
from ai_hydra.utils.HydraMetrics import HydraMetrics


class GameScorePlot(Widget):

    def __init__(self, metrics: HydraMetrics, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.metrics = metrics

    def compose(self) -> ComposeResult:
        yield Horizontal(
            PlotWidget(id=DField.PLOT_HIGHSCORES),
            PlotWidget(id=DField.PLOT_CUR_SCORE),
        )

    def clear(self):
        self.query_one(f"#{DField.PLOT_HIGHSCORES}", PlotWidget).clear()
        self.query_one(f"#{DField.PLOT_CUR_SCORE}", PlotWidget).clear()

    def plot_cur_scores(self):
        score_events = self.metrics.get_cur_score_plot_points()
        if not score_events:
            return

        episodes, scores = zip(*score_events)

        plot = self.query_one(f"#{DField.PLOT_CUR_SCORE}", PlotWidget)
        plot.clear()

        plot.plot(
            x=episodes,
            y=scores,
            line_style=DColor.GREEN,
            hires_mode=HiResMode.BRAILLE,
            label=DLabel.SCORE,
        )

        avg_score_events = self.metrics.get_avg_cur_score_plot_points()
        if avg_score_events:
            avg_episodes, avg_scores = zip(*avg_score_events)
            plot.plot(
                x=avg_episodes,
                y=avg_scores,
                line_style=DColor.RED,
                hires_mode=HiResMode.BRAILLE,
                label=DLabel.AVERAGE,
            )

        plot.set_xlabel(DLabel.EPISODES)
        plot.set_ylabel(DLabel.CUR_SCORES)
        plot.show_legend(location=LegendLocation.TOPLEFT)

    def plot_highscores(self):
        plot_points = self.metrics.get_highscore_plot_points()
        if not plot_points:
            return
        episodes, scores = zip(*plot_points)

        plot = self.query_one(f"#{DField.PLOT_HIGHSCORES}", PlotWidget)
        plot.clear()

        plot.plot(
            x=episodes,
            y=scores,
            line_style=DColor.GREEN,
            hires_mode=HiResMode.BRAILLE,
            label=DLabel.HIGHSCORE,
        )
        plot.set_xlabel(DLabel.EPISODES)
        plot.set_ylabel(DLabel.HIGHSCORES)
        plot.show_legend(location=LegendLocation.TOPLEFT)



# ai_hydra/nnet/ATH/ATHGearBox.py
#
#    AI Hydra
#    Author: Nadim-Daniel Ghaznavi
#    Copyright: (c) 2025-2026 Nadim-Daniel Ghaznavi
#    GitHub: https://github.com/NadimGhaznavi/ai_hydra
#    Website: https://ai-hydra.readthedocs.io/en/latest
#    License: GPL 3.0


from __future__ import annotations

from random import Random
from dataclasses import dataclass, field

from ai_hydra.constants.DReplayMemory import DMemDef
from ai_hydra.constants.DHydra import DHydraLog, DModule
from ai_hydra.constants.DHydraTui import DField
from ai_hydra.constants.DEvent import EV_TYPE, EV_STATUS

from ai_hydra.utils.HydraLog import HydraLog
from ai_hydra.zmq.HydraEventMQ import HydraEventMQ, EventMsg
from ai_hydra.nnet.ATH.ATHDataStore import ATHDataStore
from ai_hydra.nnet.ATH.ATHCommon import get_gear_data

MAX_BUCKETS = DMemDef.MAX_BUCKETS
THRESHOLD_BUCKETS = DMemDef.THRESHOLD_BUCKETS
UPSHIFT_COUNT_THRESHOLD = DMemDef.UPSHIFT_COUNT_THRESHOLD
DOWNSHIFT_COUNT_THRESHOLD = DMemDef.DOWNSHIFT_COUNT_THRESHOLD
NUM_COOLDOWN_EPISODES = DMemDef.NUM_COOLDOWN_EPISODES


class ATHGearBox:

    def __init__(
        self, log_level: DHydraLog, data_store: ATHDataStore, pub_func
    ):
        self.store = data_store

        # Local logging
        self.log = HydraLog(
            client_id=DModule.ATH_GEARBOX,
            log_level=log_level,
            to_console=True,
        )
        # Remote logging: ZeroMQ "Events" PUB/SUB topic
        self.event = HydraEventMQ(
            client_id=DModule.ATH_GEARBOX, pub_func=pub_func
        )

        self._cur_gear = 1
        self._cur_seq_length, self._cur_batch_size = get_gear_data(
            self._cur_gear
        )

        self._cur_bucket_counts: int | None = None

        self._cooldown_count = 0

        # Handling stagnation by downshifting
        self._stagnation_flag = False
        self._stagnation_alert_count = 0
        # Track critical stagnation events
        self._crit_stagnation_alert_count = 0

        # Track when the last upshift happened
        self._cur_epoch = 0
        self._last_upshift = 0

    async def get_gear(self) -> int:
        self._cur_bucket_counts = bucket_counts = (
            self.store.get_bucket_counts()
        )

        if not bucket_counts:
            return self._cur_gear

        if len(bucket_counts) != MAX_BUCKETS:
            return self._cur_gear

        chunk_count = self.get_thresh_bucket_chunk_count()

        # We're dealing with a stagnation event
        if (
            self._stagnation_flag
            and self._cooldown_count > NUM_COOLDOWN_EPISODES
        ):

            # Downshift
            self._cur_gear -= self._stagnation_alert_count

            # Make sure we don't downshift past 1
            if self._cur_gear < 1:
                self._cur_gear = 1
                return self._cur_gear

            self._cur_seq_length, self._cur_batch_size = get_gear_data(
                self._cur_gear
            )
            # Stay here till the cooldown period is over
            self._cooldown_count = 0
            # Clear the stagnation flag
            self._stagnation_flag = False

            # Log the event
            await self.event.publish(
                EventMsg(
                    level=EV_STATUS.WARN,
                    message=(
                        f"Stagnation alert({self._stagnation_alert_count}) - "
                        f"Shifting DOWN: {self._cur_gear + self._stagnation_alert_count}"
                        f" > {self._cur_gear} - {self._cur_seq_length}/{self._cur_batch_size}"
                    ),
                    ev_type=EV_TYPE.SHIFTING,
                    payload={
                        DField.GEAR: self._cur_gear,
                        DField.SEQ_LENGTH: self._cur_seq_length,
                        DField.BATCH_SIZE: self._cur_batch_size,
                    },
                )
            )

        # Time to shift UP
        elif (
            chunk_count > UPSHIFT_COUNT_THRESHOLD
            and self._cooldown_count > NUM_COOLDOWN_EPISODES
        ):
            # Upshifting resets the stagnation alert count. This allows the
            # system to, up, up, stagnation-alert: down, up, up, ... to
            # work it's way back up to the "top gear" again.
            self._stagnation_alert_count = 0

            self._cur_gear += 1
            self._last_upshift = self._cur_epoch
            self._cooldown_count = 0
            self._cur_seq_length, self._cur_batch_size = get_gear_data(
                self._cur_gear
            )
            await self.event.publish(
                EventMsg(
                    level=EV_STATUS.GOOD,
                    message=f"Shifting UP: {self._cur_gear - 1} > {self._cur_gear} - {self._cur_seq_length}/{self._cur_batch_size}",
                    ev_type=EV_TYPE.SHIFTING,
                    payload={
                        DField.GEAR: self._cur_gear,
                        DField.SEQ_LENGTH: self._cur_seq_length,
                        DField.BATCH_SIZE: self._cur_batch_size,
                    },
                )
            )

        # Time to shift DOWN
        elif (
            chunk_count < DOWNSHIFT_COUNT_THRESHOLD
            and self._cooldown_count > NUM_COOLDOWN_EPISODES
        ):
            # First gear is the lowest gear
            if self._cur_gear == 1:
                return self._cur_gear

            self._cur_gear -= 1
            self._cooldown_count = 0
            self._cur_seq_length, self._cur_batch_size = get_gear_data(
                self._cur_gear
            )
            await self.event.publish(
                EventMsg(
                    level=EV_STATUS.WARN,
                    message=f"Shifting DOWN: {self._cur_gear} - {self._cur_seq_length}/{self._cur_batch_size}",
                    ev_type=EV_TYPE.SHIFTING,
                    payload={
                        DField.GEAR: self._cur_gear,
                        DField.SEQ_LENGTH: self._cur_seq_length,
                        DField.BATCH_SIZE: self._cur_batch_size,
                    },
                )
            )

        return self._cur_gear

    def get_thresh_bucket_chunk_count(self) -> int:
        count = 0
        for bucket_idx in THRESHOLD_BUCKETS:
            count += self._cur_bucket_counts[bucket_idx]
        return count

    async def hard_reset(self, crit_count: int):
        self._stagnation_flag = False
        self._cooldown_count = 0
        old_gear = self._cur_gear
        self._cur_gear = 3  # seq_length/batch_size == 8/32
        self._cur_seq_length, self._cur_batch_size = get_gear_data(
            self._cur_gear
        )
        msg = (
            f"Critical Stagnation alert({crit_count}): Radical DOWN shift: "
            f"{old_gear} > {self._cur_gear} - {self._cur_seq_length}/"
            f"{self._cur_batch_size}"
        )
        await self.event.publish(
            EventMsg(
                level=EV_STATUS.BAD,
                message=msg,
                ev_type=EV_TYPE.SHIFTING,
                payload={
                    DField.GEAR: self._cur_gear,
                    DField.SEQ_LENGTH: self._cur_seq_length,
                    DField.BATCH_SIZE: self._cur_batch_size,
                },
            )
        )
        self.log.debug(msg)
        return

    def incr_cooldown_count(self):
        self._cooldown_count += 1

    def incr_epoch(self):
        self._cur_epoch += 1

    async def stagnation_cleared(self):
        if self._stagnation_alert_count != 0:
            msg = "Resetting stagnation count to 0"
            await self.event.publish(
                EventMsg(
                    level=EV_STATUS.INFO,
                    message=msg,
                )
            )
            self.log.debug(msg)
        self._stagnation_flag = False
        self._stagnation_alert_count = 0

    def stagnation_warning(self):
        # Don't downshift if we just upshifted....
        num_ep_since_upshift = self._cur_epoch - self._last_upshift
        if num_ep_since_upshift > DMemDef.MAX_STAGNANT_EPISODES:
            self.log.debug(
                "Received stagnation warning: Setting stagnation flag"
            )
            self._stagnation_flag = True
            self._stagnation_alert_count += 1
            return

        self.log.debug(
            "Received stagnation warning: Ignoring because of recent upshift"
        )



# ai_hydra/utils/HydraClientMQ.py
#
#    AI Hydra
#    Author: Nadim-Daniel Ghaznavi
#    Copyright: (c) 2025-2026 Nadim-Daniel Ghaznavi
#    GitHub: https://github.com/NadimGhaznavi/ai_hydra
#    Website: https://ai-hydra.readthedocs.io/en/latest
#    License: GPL 3.0
#
"""
_init_sub_socket()
connect_sub()
subscribe()
unsubscribe()
bg_sub_listen()
"""

from typing import Any
from collections.abc import Awaitable, Callable
import asyncio
import zmq, zmq.asyncio
import json
import traceback

from ai_hydra.constants.DHydra import (
    DHydra,
    DHydraRouterDef,
    DModule,
    DHydraServerDef,
)
from ai_hydra.constants.DHydraMQ import DHydraMQDef, DHydraMQ
from ai_hydra.zmq.HydraBaseMQ import HydraBaseMQ
from ai_hydra.zmq.HydraMsg import HydraMsg

SubHandler = Callable[[str, dict], Any | Awaitable[Any]]


class HydraClientMQ(HydraBaseMQ):
    def __init__(
        self,
        *,
        router_address: str = DHydraRouterDef.HOSTNAME,
        router_port: int = DHydraRouterDef.PORT,
        router_hb_port: int = DHydraRouterDef.HEARTBEAT_PORT,
        identity: str = DModule.HYDRA_MQ,
        topic_prefix: str = DHydraMQDef.TOPIC_PREFIX,
        srv_host: str = DHydraServerDef.HOSTNAME,
        sub_port: int = DHydraServerDef.PUB_PORT,
        sub_methods: dict[str, SubHandler] | None = None,
    ) -> None:

        super().__init__(
            router_address=router_address,
            router_port=router_port,
            router_hb_port=router_hb_port,
            identity=identity,
            topic_prefix=topic_prefix,
        )
        self.srv_host = srv_host
        self.sub_port = sub_port
        self.sub_methods = sub_methods or {}

        self.sub_socket: zmq.asyncio.Socket | None = None
        self.sub_task: asyncio.Task[None] | None = None
        self.sub_addr = f"tcp://{srv_host}:{sub_port}"
        self.sub_socket = self.ctx.socket(zmq.SUB)

        self.sub_socket.connect(self.sub_addr)
        self.sub_stop_event = asyncio.Event()

    async def bg_sub_listen(self) -> None:
        """
        Background subscriber loop for telemetry.
        """
        if self.sub_socket is None:
            return

        try:
            while not self.sub_stop_event.is_set():
                try:
                    frames = await asyncio.wait_for(
                        self.sub_socket.recv_multipart(copy=True),
                        timeout=DHydra.NETWORK_TIMEOUT,
                    )

                    if len(frames) != 2:
                        print(
                            f"ERROR: telemetry expected 2 frames, got {len(frames)}"
                        )
                        continue

                    topic = self._ensure_bytes(frames[0]).decode(
                        DHydraMQ.UTF_8, errors="replace"
                    )
                    payload_bytes = self._ensure_bytes(frames[1])

                    try:
                        payload = json.loads(payload_bytes.decode("utf-8"))
                    except Exception as e:
                        print(f"ERROR: SUB JSON decode error: {e}")
                        continue

                    handler = self.sub_methods.get(topic)

                    if handler is not None:
                        result = handler(topic, payload)
                        if asyncio.iscoroutine(result):
                            await result

                except asyncio.TimeoutError:
                    pass
                except Exception as e:
                    print(f"ERROR: telemetry listen error: {e}")
                    print(f"STACKTRACK: {traceback.format_exc()}")

        except asyncio.CancelledError:
            raise

    def disable_events_sub(self) -> None:
        self._sub_set(False, self.topic(DHydraMQDef.EVENTS_TOPIC))

    def disable_per_step_sub(self) -> None:
        self._sub_set(False, self.topic(DHydraMQDef.PER_STEP_TOPIC))

    def disable_per_episode_sub(self) -> None:
        self._sub_set(False, self.topic(DHydraMQDef.PER_EPISODE_TOPIC))

    def disable_scores_sub(self) -> None:
        self._sub_set(False, self.topic(DHydraMQDef.SCORES_TOPIC))

    def enable_events_sub(self) -> None:
        self._sub_set(True, self.topic(DHydraMQDef.EVENTS_TOPIC))

    def enable_per_step_sub(self) -> None:
        self._sub_set(True, self.topic(DHydraMQDef.PER_STEP_TOPIC))

    def enable_per_episode_sub(self) -> None:
        self._sub_set(True, self.topic(DHydraMQDef.PER_EPISODE_TOPIC))

    def enable_scores_sub(self) -> None:
        self._sub_set(True, self.topic(DHydraMQDef.SCORES_TOPIC))

    async def quit(self) -> None:
        await super().quit()
        if self._stopped:
            return
        self._stopped = True
        self._started = False

        # 1. stop all tasks first
        self.hb_stop_event.set()
        self.sub_stop_event.set()

        if self.hb_task is not None:
            self.hb_task.cancel()
            try:
                await self.hb_task
            except asyncio.CancelledError:
                pass
            finally:
                self.hb_task = None

        if self.sub_task is not None:
            self.sub_task.cancel()
            try:
                await self.sub_task
            except asyncio.CancelledError:
                pass
            finally:
                self.sub_task = None

        # 2. close sockets
        self._ignore_zmq_teardown(
            lambda: self.socket.disconnect(self.router_addr),
            f"socket.disconnect({self.router_addr})",
        )
        self._ignore_zmq_teardown(
            lambda: self.socket.close(linger=0),
            "socket.close(linger=0)",
        )

        self._ignore_zmq_teardown(
            lambda: self.hb_socket.disconnect(self.router_hb_addr),
            f"hb_socket.disconnect({self.router_hb_addr})",
        )
        self._ignore_zmq_teardown(
            lambda: self.hb_socket.close(linger=0),
            "hb_socket.close(linger=0)",
        )

        if self.sub_socket is not None:
            self._ignore_zmq_teardown(
                lambda: self.sub_socket.close(linger=0),
                "sub_socket.close(linger=0)",
            )
            self.sub_socket = None

        # 3. terminate context once
        try:
            self.ctx.term()
        except zmq.ZMQError as e:
            print(f"DEBUG: ignoring ctx.term() during shutdown: {e}")

    def start(self) -> None:
        super().start()

        if self.sub_task is None:
            self.sub_task = asyncio.create_task(
                self.bg_sub_listen(), name="sub_listen"
            )

    def _sub_set(self, subscribe: bool, prefix: str) -> None:
        if self.sub_socket is None:
            raise RuntimeError("SUB not configured")
        opt = zmq.SUBSCRIBE if subscribe else zmq.UNSUBSCRIBE
        self.sub_socket.setsockopt(opt, prefix.encode(DHydraMQ.UTF_8))



# ai_hydra/nnet/LinearTrainer.py
#
#    AI Hydra
#    Author: Nadim-Daniel Ghaznavi
#    Copyright: (c) 2025-2026 Nadim-Daniel Ghaznavi
#    GitHub: https://github.com/NadimGhaznavi/ai_hydra
#    Website: https://ai-hydra.readthedocs.io/en/latest
#    License: GPL 3.0

import torch
import torch.nn as nn
import torch.optim as optim

from ai_hydra.constants.DHydraTui import DField
from ai_hydra.constants.DHydra import DHydra, DHydraLog, DModule

from ai_hydra.nnet.SimpleReplayMemory import SimpleReplayMemory
from ai_hydra.utils.HydraLog import HydraLog


class LinearTrainer:
    def __init__(
        self,
        model,
        replay: SimpleReplayMemory,
        lr: float,
        log_level: DHydraLog,
        device: torch.device | None = None,
    ):
        self.device = device or torch.device(DField.CPU)
        self.model = model.to(self.device)
        self.replay = replay
        self._gamma = None
        self._batch_size = None

        self.optimizer = optim.Adam(self.model.parameters(), lr=lr)
        self.criterion = nn.SmoothL1Loss()  # nn.MSELoss()
        torch.manual_seed(DHydra.RANDOM_SEED)
        self.log = HydraLog(
            client_id=DModule.LINEAR_TRAINER,
            log_level=log_level,
            to_console=True,
        )
        self._losses = []
        self._cold_memory = True
        self.log.info(f"Setting the learning rate to {lr}")
        self.log.info("Initialized")

    def get_avg_loss(self) -> float | None:
        if not self._losses:
            return None
        avg_loss = sum(self._losses) / len(self._losses)
        self._losses = []
        return avg_loss

    def reset(self):
        self.optimizer.state = {}

    async def train_long_memory(self) -> float | None:

        batch = await self.replay.sample_transitions(self._batch_size)
        if batch is None:
            return None

        if self._cold_memory:
            self._cold_memory = False
            self.log.debug(f"Training with {self._batch_size} transitions")

        states = torch.tensor(
            [t.old_state for t in batch],
            dtype=torch.float32,
            device=self.device,
        )
        actions = torch.tensor(
            [t.action for t in batch], dtype=torch.int64, device=self.device
        )
        rewards = torch.tensor(
            [t.reward for t in batch], dtype=torch.float32, device=self.device
        )
        next_states = torch.tensor(
            [t.new_state for t in batch],
            dtype=torch.float32,
            device=self.device,
        )
        dones = torch.tensor(
            [t.done for t in batch], dtype=torch.float32, device=self.device
        )

        self.model.train()

        q_values = self.model(states)
        q_pred = q_values.gather(1, actions.unsqueeze(1)).squeeze(1)

        self.model.eval()
        with torch.no_grad():
            next_q = self.model(next_states)
            max_next_q = next_q.max(dim=1).values
            q_target = rewards + self._gamma * max_next_q * (1.0 - dones)

        loss = self.criterion(q_pred, q_target)

        self.optimizer.zero_grad()
        loss.backward()
        self.optimizer.step()

        self._losses.append(loss.item())

        return float(loss.item())

    def set_params(self, gamma: float, batch_size: int):
        self._gamma = gamma
        self.log.info(f"Setting the Discount/Gamma to {gamma}")
        self._batch_size = batch_size
        self.log.info(f"Setting the Batch Size to {batch_size}")



# ai_hydra/nnet/ATH/ATHMemory.py
#
#    AI Hydra
#    Author: Nadim-Daniel Ghaznavi
#    Copyright: (c) 2025-2026 Nadim-Daniel Ghaznavi
#    GitHub: https://github.com/NadimGhaznavi/ai_hydra
#    Website: https://ai-hydra.readthedocs.io/en/latest
#    License: GPL 3.0

from random import Random

from ai_hydra.constants.DHydra import DHydraLog, DModule

from ai_hydra.utils.HydraLog import HydraLog
from ai_hydra.nnet.Transition import Transition
from ai_hydra.nnet.ATH.ATHGearbox import ATHGearBox
from ai_hydra.nnet.ATH.ATHDataStore import ATHDataStore
from ai_hydra.nnet.ATH.ATHDataMgr import ATHDataMgr


class ATHMemory:
    def __init__(self, rng: Random, log_level: DHydraLog, pub_func):
        # Local logging
        self.log = HydraLog(
            client_id=DModule.ATH_MEMORY,
            log_level=log_level,
            to_console=True,
        )

        # Memory storage
        data_store = ATHDataStore(log_level=log_level, pub_func=pub_func)

        # Memory Management
        self.data_mgr = ATHDataMgr(
            log_level=log_level,
            rng=rng,
            pub_func=pub_func,
            data_store=data_store,
        )

        # The ATH Gearbox a.k.a. "The Automatic ATH Transmission."
        self.gearbox = ATHGearBox(
            log_level=log_level, pub_func=pub_func, data_store=data_store
        )

    async def append(
        self, t: Transition, final_score: int | None = None
    ) -> None:
        if t.done:
            self.gearbox.incr_cooldown_count()
        await self.data_mgr.append(t=t, final_score=final_score)
        self.data_mgr.set_gear(await self.gearbox.get_gear())



# ai_hydra/nnet/ATHStore.py
#
#    AI Hydra
#    Author: Nadim-Daniel Ghaznavi
#    Copyright: (c) 2025-2026 Nadim-Daniel Ghaznavi
#    GitHub: https://github.com/NadimGhaznavi/ai_hydra
#    Website: https://ai-hydra.readthedocs.io/en/latest
#    License: GPL 3.0


from __future__ import annotations

from random import Random


from ai_hydra.constants.DReplayMemory import DMemDef, DMemField
from ai_hydra.constants.DHydra import DHydraLog, DModule
from ai_hydra.constants.DHydraTui import DField
from ai_hydra.constants.DEvent import EV_TYPE

from ai_hydra.nnet.Transition import Transition
from ai_hydra.utils.HydraLog import HydraLog
from ai_hydra.zmq.HydraEventMQ import HydraEventMQ, EventMsg

from ai_hydra.nnet.ATH.ATHCommon import Episode, GearMeta, get_gear_data

MAX_BUCKETS = DMemDef.MAX_BUCKETS


class ATHDataStore:
    def __init__(self, log_level: DHydraLog, pub_func):

        # Local logging
        self.log = HydraLog(
            client_id=DModule.ATH_DATA_STORE,
            log_level=log_level,
            to_console=True,
        )
        # Remote logging: ZeroMQ "Events" PUB/SUB topic
        self.event = HydraEventMQ(
            client_id=DModule.ATH_DATA_STORE, pub_func=pub_func
        )

        self._episodes_by_bucket: dict[int, list[Episode]] = {
            idx: [] for idx in range(MAX_BUCKETS)
        }
        self._warmed_buckets: list[int] = []
        self._games: list[Episode] = []
        self._cur_gear: int | None = None
        self._cur_game = []
        self._batch_size: int | None = None

        self._stored_frames = 0

    def append(self, t: Transition):
        self._cur_game.append(t)

    def finalize_game(self, gear_meta: dict[int, GearMeta]) -> None:
        frames = tuple(self._cur_game)
        self._cur_game = []
        ep_size = len(frames)
        self._stored_frames += ep_size

        self._games.append(
            Episode(
                frames=frames,
                size=ep_size,
                gear_meta=gear_meta,
            )
        )

    def get_cur_ep_size(self) -> int:
        return len(self._cur_game)

    def get_bucket_counts(self) -> dict[int, int]:
        return {
            idx: len(episodes)
            for idx, episodes in self._episodes_by_bucket.items()
            if episodes
        }

    def get_eps_by_bucket_idx(self, bucket_idx: int):
        return self._episodes_by_bucket[bucket_idx]

    def get_stored_frame_count(self) -> int | None:
        return self._stored_frames

    def get_warmed_buckets(self):
        return self._warmed_buckets

    def pop_first_game(self):
        episode = self._games.pop(0)
        self._stored_frames -= episode.size

    def update_warmed_buckets_list(self):
        self._warmed_buckets = [
            idx
            for idx, episodes in self._episodes_by_bucket.items()
            if episodes
        ]

    def update_episodes_by_bucket(self):
        for ep in self._games:
            meta = ep.gear_meta[self._cur_gear]
            usable_buckets = min(meta.num_chunks, MAX_BUCKETS)

            for bucket_idx in range(usable_buckets):
                self._episodes_by_bucket[bucket_idx].append(ep)

    def reset_bucket_indexes(self):
        self._episodes_by_bucket = {idx: [] for idx in range(MAX_BUCKETS)}

    def set_gear(self, gear: int):
        self._cur_gear = gear
        _, self._batch_size = get_gear_data(gear)



# ai_hydra/constants/DHydraTui.py
#
#    AI Hydra
#    Author: Nadim-Daniel Ghaznavi
#    Copyright: (c) 2025-2026 Nadim-Daniel Ghaznavi
#    GitHub: https://github.com/NadimGhaznavi/ai_hydra
#    Website: https://ai-hydra.readthedocs.io/en/latest
#    License: GPL 3.0

from typing import Final


class DColor:
    """Color constants"""

    BLACK: Final[str] = "black"
    BLUE: Final[str] = "blue"
    DARK_RED: Final[str] = "dark_red"
    GREEN: Final[str] = "green"
    PURPLE: Final[str] = "purple"
    RED: Final[str] = "red"
    WHITE: Final[str] = "white"
    YELLOW: Final[str] = "yellow"


class DField:
    """
    Machine readable field names.
    """

    ALL: Final[str] = "all"
    ATH_Memory: Final[str] = "ath_memory"
    BAD_HANDSHAKE: Final[str] = "bad_handshake"
    BATCH_SIZE: Final[str] = "batch_size"
    BATCH_SIZE_INPUT: Final[str] = "batch_size_input"
    BATCH_SIZE_LABEL: Final[str] = "batch_size_label"
    BOARD_BOX: Final[str] = "board_box"
    BOTH: Final[str] = "both"
    BUTTONS: Final[str] = "buttons"
    CLIENTS: Final[str] = "clients"
    CLIENTS_SCREEN: Final[str] = "clients_screen"
    CONFIG: Final[str] = "config"
    CONSOLE_BOX: Final[str] = "console_box"
    CPU: Final[str] = "cpu"
    DISCONNECTED: Final[str] = "disconnected"
    CONSOLE: Final[str] = "console"
    CONSOLE_SCREEN: Final[str] = "console_screen"
    COUNT: Final[str] = "count"
    CUR_EPSILON: Final[str] = "cur_epsilon"
    CUR_LOSS: Final[str] = "cur_loss"
    DROPOUT_P: Final[str] = "dropout_p"
    DROPOUT_P_INPUT: Final[str] = "dropout_p_input"
    DROPOUT_P_LABEL: Final[str] = "dropout_p_label"
    EPSILON: Final[str] = "epsilon"
    EPSILON_DECAY: Final[str] = "epsilon_decay"
    EPSILON_DECAY_INPUT: Final[str] = "epsilon_decay_input"
    EPSILON_DECAY_LABEL: Final[str] = "epsilon_decay_label"
    EPSILON_DEPLETED: Final[str] = "epsilon_depleted"
    EVENT_LOG: Final[str] = "event_log"
    EVENT_LOG_LOG: Final[str] = "event_log_log"
    EVENTS: Final[str] = "events"
    GEAR: Final[str] = "gear"
    GAMMA_INPUT: Final[str] = "gamma_input"
    GAMMA_LABEL: Final[str] = "gamma_label"
    GAME_SCORE_PLOT: Final[str] = "game_score_plot"
    GAME_SCORES: Final[str] = "game_scores"
    GRU: Final[str] = "gru"
    HANDSHAKE: Final[str] = "handshake"
    HANDSHAKE_DIVIDER: Final[str] = "handshake_divider"
    HIDDEN_SIZE_INPUT: Final[str] = "hidden_size_input"
    HIDDEN_SIZE_LABEL: Final[str] = "hidden_size_label"
    HIDDEN_WIDGET: Final[str] = "hidden_widget"
    HIGHSCORE: Final[str] = "highscore"
    HIGHSCORES: Final[str] = "highscores"
    HIGHSCORES_BOX: Final[str] = "highscores_box"
    HIGHSCORES_LOG: Final[str] = "highscores_log"
    HIGHSCORES_PLOT: Final[str] = "highscores_plot"
    HIGHSCORES_LH: Final[str] = "highscores_lh"
    HYDRA_TELEMETRY: Final[str] = "hydra_telemetry"
    INITIAL_EPSILON: Final[str] = "initial_epsilon"
    INITIAL_EPSILON_INPUT: Final[str] = "initial_epsilon_input"
    INITIAL_EPSILON_LABEL: Final[str] = "initial_epsilon_label"
    INPUT_FIELD: Final[str] = "input_field"
    INPUT_SIZE: Final[str] = "input_size"
    INTEGER: Final[str] = "integer"
    LEARNING_RATE_INPUT: Final[str] = "learning_rate_input"
    LEARNING_RATE_LABEL: Final[str] = "learning_rate_label"
    LINEAR: Final[str] = "linear"
    LOOKAHEAD_BOX: Final[str] = "lookahead_box"
    LOOKAHEAD_STATUS: Final[str] = "lookahead_status"
    LOOKAHEAD_P_VAL_INPUT: Final[str] = "lookahead_p_val_input"
    LOOKAHEAD_P_VAL_LABEL: Final[str] = "lookahead_p_val_label"
    LOOKAHEAD_SAMPLE_P_VALUE_LABEL: Final[str] = "lookahead_sample_p_val_label"
    LOSS: Final[str] = "loss"
    LOSS_PLOT: Final[str] = "loss_plot"
    MIN_EPSILON: Final[str] = "min_epsilon"
    MIN_EPSILON_INPUT: Final[str] = "min_epsilon_input"
    MIN_EPSILON_LABEL: Final[str] = "min_epsilon_label"
    MODEL: Final[str] = "model"
    MODEL_TYPE_LABEL: Final[str] = "model_type_label"
    MODEL_TYPE_SELECT: Final[str] = "model_type_select"
    MOVE_DELAY_LABEL: Final[str] = "move_delay_label"
    MOVE_DELAY_INPUT: Final[str] = "move_delay_input"
    NETWORK: Final[str] = "network"
    NORMAL: Final[str] = "normal"
    NUMBER: Final[str] = "number"
    PAUSE_RUN: Final[str] = "pause_run"
    PAUSE_RUN_DIVIDER: Final[str] = "pause_run_divider"
    PLOT_CUR_SCORE: Final[str] = "plot_cur_score"
    PLOT_HIGHSCORES: Final[str] = "plot_highscores"
    PLOT_LOSS: Final[str] = "plot_loss"
    PLOT_RECENT_LOSS: Final[str] = "plot_recent_loss"
    PLOT_RECENT_SCORES_DIST: Final[str] = "plot_recent_scores_dist"
    PLOT_SCORES_DIST: Final[str] = "plot_scores_dist"
    QUIT: Final[str] = "quit"
    RECENT: Final[str] = "recent"
    REPLAY_MEM: Final[str] = "replay_mem"
    RANDOM_SEED_INPUT: Final[str] = "random_seed_input"
    RANDOM_SEED_LABEL: Final[str] = "random_seed_label"
    RESET: Final[str] = "reset"
    RESET_RUN: Final[str] = "reset_run"
    RESET_RUN_DIVIDER: Final[str] = "reset_run_divider"
    RESUME_RUN: Final[str] = "resume_run"
    RESUME_RUN_DIVIDER: Final[str] = "resume_run_divider"
    RNN: Final[str] = "rnn"
    RNN_BATCH_SIZE_LABEL: Final[str] = "rnn_batch_size_label"
    RNN_TAU_INPUT: Final[str] = "rnn_tau_input"
    RNN_TAU_LABEL: Final[str] = "rnn_tau_label"
    RNN_TAU_OPT: Final[str] = "rnn_tau_opt"
    RNN_LAYERS_INPUT: Final[str] = "rnn_layers_input"
    RNN_LAYERS_LABEL: Final[str] = "rnn_layers_label"
    RNN_LAYERS_OPT: Final[str] = "rnn_layers_opt"
    RNN_DROPOUT: Final[str] = "rnn_dropout"
    RNN_LAYERS: Final[str] = "rnn_layers"
    ROUTER_HB: Final[str] = "router_hb"
    RUNNING: Final[str] = "running"
    SCORES_DIST_PLOT: Final[str] = "scores_list_plot"
    SEQ_LENGTH: Final[str] = "seq_length"
    SETTINGS: Final[str] = "settings"
    SEQ_LENGTH_INPUT: Final[str] = "seq_length_input"
    SEQ_LENGTH_LABEL: Final[str] = "seq_length_label"
    SEQ_LENGTH_OPT: Final[str] = "seq_length_opt"
    SCORES_PLOT: Final[str] = "scores_plot"
    SCORES_PLOT_LH: Final[str] = "scores_plot_lh"
    SCORES_PLOT_NLH: Final[str] = "scores_plot_nlh"
    SCORES_SCATTER_PLOT: Final[str] = "scores_scatter_plot"
    SIM_LOOP: Final[str] = "sim_loop"
    SIM_PAUSED: Final[str] = "sim_paused"
    SIM_RUNNING: Final[str] = "sim_running"
    SIM_STOPPED: Final[str] = "sim_stopped"
    SNAPSHOT: Final[str] = "snapshot"
    SNAPSHOT_DIVIDER: Final[str] = "snapshot_divider"
    STATUS: Final[str] = "status"
    START: Final[str] = "start"
    START_RUN: Final[str] = "start_run"
    START_RUN_DIVIDER: Final[str] = "start_run_divider"
    STOP: Final[str] = "stop"
    STOPPED: Final[str] = "stopped"
    TABBED_PLOTS: Final[str] = "tabbed_plots"
    TANH: Final[str] = "tanh"
    TITLE: Final[str] = "title"
    TRAINING_BOX: Final[str] = "training_box"
    TUI: Final[str] = "tui"
    TURBO_BOX: Final[str] = "turbo_box"
    TURBO_MODE: Final[str] = "turbo_mode"
    TURBO_OFF: Final[str] = "turbo_off"
    TURBO_ON: Final[str] = "turbo_on"
    UPDATE_RUNTIME_CONFIG: Final[str] = "update_runtime_config"
    UPDATE_CONFIG_SPACER: Final[str] = "update_config_spacer"
    WARNING: Final[str] = "warning"


class DFile:
    """
    Filenames.
    """

    CLIENT_CSS: Final[str] = "HydraClient.tcss"
    ROUTER_CSS: Final[str] = "HydraRouter.tcss"
    HYDRA_SERVER_DB: Final[str] = "HydraServer.db"
    BASE_SNAPSHOT: Final[str] = "AI-Hydra-Snapshot_"


class DLabel:
    """
    Human readable text.
    """

    ACTIONS: Final[str] = "Actions"
    AVERAGE: Final[str] = "Average"
    BATCH_SIZE: Final[str] = "Batch Size"
    BOTH: Final[str] = "Both"
    CLIENT_TITLE: Final[str] = "Hydra Client"
    CLIENTS: Final[str] = "Clients"
    COMPLETE: Final[str] = "Complete"
    CONFIG: Final[str] = "Configuration"
    CONSOLE: Final[str] = "Console"
    CUR_LOSS: Final[str] = "Current Loss"
    CUR_EPSILON: Final[str] = "Current Epsilon"
    CURRENT: Final[str] = "Current"
    CUR_SCORES: Final[str] = "Current Scores"
    COUNT: Final[str] = "Count"
    DEBUG: Final[str] = "DEBUG"
    DISCONNECTED: Final[str] = "Disonnected"
    DROPOUT_P_VAL: Final[str] = "Dropout P-Value"
    EPISODE: Final[str] = "Episode"
    EPISODES: Final[str] = "Episodes"
    EPSILON: Final[str] = "Epsilon"
    EPSILON_DECAY: Final[str] = "Epsilon Decay"
    ERROR: Final[str] = "ERROR"
    EVENT: Final[str] = "Event"
    EVENT_LOG: Final[str] = "Event Log"
    EVENT_TIME: Final[str] = "Event"
    GAME: Final[str] = "Game"
    GAME_SCORES: Final[str] = "Game Scores"
    GAMMA: Final[str] = "Discount/Gamma"
    GRU: Final[str] = "GRU"
    HANDSHAKE: Final[str] = "Handshake"
    HIDDEN_SIZE: Final[str] = "Hidden Size"
    HIGHSCORE: Final[str] = "Highscore"
    HIGHSCORES: Final[str] = "Highscores"
    HIGHSCORES_LH: Final[str] = "Look Ahead"
    INITIAL_EPSILON: Final[str] = "Initial Epsilon"
    LEFT: Final[str] = "Left"
    LH: Final[str] = "LA"
    LEARNING_RATE: Final[str] = "Learning Rate"
    LINEAR: Final[str] = "Linear"
    LISTEN_PORT: Final[str] = "Listening Port"
    LOOKAHEAD: Final[str] = "Look Ahead"
    LOOKAHEAD_SETTINGS: Final[str] = "Look Ahead Settings"
    LOSS: Final[str] = "Loss"
    MEAN: Final[str] = "Mean"
    MEDIAN: Final[str] = "Median"
    MIN_EPSILON: Final[str] = "Minimum Epsilon"
    MODEL: Final[str] = "Model"
    MOVE_DELAY: Final[str] = "Move Delay"
    NETWORK: Final[str] = "Network"
    NN_MODEL: Final[str] = "NN Model"
    NORMAL: Final[str] = "Normal"
    P_VALUE: Final[str] = "P-Value"
    PAUSE: Final[str] = "Pause"
    PING_ROUTER: Final[str] = "Ping Router"
    PING_SERVER: Final[str] = "Ping Server"
    QUIT: Final[str] = "Quit"
    RANDOM_SEED: Final[str] = "Random Seed"
    RECENT_COUNT: Final[str] = "Recent Scores"
    RECENT_LOSS: Final[str] = "Recent Loss"
    RESET: Final[str] = "Reset"
    RESUME: Final[str] = "Resume"
    RIGHT: Final[str] = "Right"
    RNN: Final[str] = "RNN"
    RNN_LAYERS: Final[str] = "Layers"
    RNN_TAU: Final[str] = "Tau"
    ROUTER: Final[str] = "Router"
    ROUTER_TITLE: Final[str] = "Hydra Router"
    SETTINGS: Final[str] = "Settings"
    SAMPLING_P_VALUE: Final[str] = "Sampling P-Value"
    SCORE: Final[str] = "Score"
    SCORE_DISTRIBUTION: Final[str] = "Score Distribution"
    SCORES: Final[str] = "Scores"
    SCORES_ALL: Final[str] = "Scores/All"
    SCORES_LH: Final[str] = "Scores/L.A."
    SCORES_NLH: Final[str] = "Scores/No L.A."
    SCORES_SCATTER: Final[str] = "Scores Scatterplot"
    SEQUENCE_LENGTH: Final[str] = "Sequence Length"
    SNAPSHOT: Final[str] = "Snapshot"
    SPACE: Final[str] = " "
    START: Final[str] = "Start"
    STATUS: Final[str] = "Status"
    STEPS: Final[str] = "Steps"
    STOP: Final[str] = "Stop"
    STATUS: Final[str] = "Status"
    STRAIGHT: Final[str] = "Straight"
    TARGET_HOST: Final[str] = "Host"
    TARGET_PORT: Final[str] = "Port"
    TIME: Final[str] = "Time"
    TRAINING: Final[str] = "Training"
    TURBO_MODE: Final[str] = "Turbo Mode"
    TYPE: Final[str] = "Type"
    UPDATE_CONFIG: Final[str] = "Update Config"
    VERSION: Final[str] = "Version"
    VISUALIZATIONS: Final[str] = "Visualizations"


class DPlotDef:
    """
    Default Textual plot settings.
    """

    # The average is calculated by dividing the MAX_DATA_POINTS by this number
    AVG_DIVISOR: int = 50

    MAX_LOSS_DATA_POINTS: Final[int] = 125
    MAX_CUR_DATA_POINTS: Final[int] = 200
    MAX_CUR_LOSS_DATA_POINTS: Final[int] = 75

    # GameScorePlot widget settings
    MAX_GAMES = 75
    MAX_GAMES_AVG_DIVISOR: int = 50

    # Maximum number of scores stored in the metrics object
    MAX_CUR_SCORES = 75

    # The number of "recent" games in the score distribution histogram
    RECENT_SCORES_MAX = 500

    # The number of "recent" loss data points in the "Recent Loss" plot
    RECENT_LOSS_MAX = 75


class DStatus:
    """
    Status indicators.
    """

    GOOD: Final[str] = "🟢"
    ON: Final[str] = "🟢"
    OK: Final[str] = "🟡"
    BAD: Final[str] = "🔴"
    OFF: Final[str] = "🔴"
    UNKNOWN: Final[str] = "⚪"



from __future__ import annotations

from dataclasses import dataclass

from ai_hydra.constants.DHydra import DModule, DHydraLog
from ai_hydra.constants.DReplayMemory import DMemDef
from ai_hydra.constants.DEvent import EV_STATUS, EV_TYPE


from ai_hydra.utils.HydraLog import HydraLog
from ai_hydra.nnet.ATH.ATHMemory import ATHMemory
from ai_hydra.nnet.LinearTrainer import LinearTrainer
from ai_hydra.nnet.Policy.HydraPolicy import HydraPolicy
from ai_hydra.nnet.Policy.EpsilonNicePolicy import EpsilonNicePolicy
from ai_hydra.nnet.RecurrentTrainer import RecurrentTrainer
from ai_hydra.nnet.SimpleReplayMemory import SimpleReplayMemory
from ai_hydra.nnet.models.LinearModel import LinearModel
from ai_hydra.nnet.models.RNNModel import RNNModel
from ai_hydra.nnet.models.GRUModel import GRUModel
from ai_hydra.server.SnakeMgr import SnakeMgr
from ai_hydra.zmq.HydraEventMQ import EventMsg, HydraEventMQ


@dataclass(frozen=True)
class EpisodeStats:
    episode_id: int
    steps: int
    score: int
    reward_total: int
    losses: int
    loss_avg: float


@dataclass(frozen=True)
class TrainRunStats:
    episodes: int
    avg_score: float
    avg_steps: float
    avg_reward_total: float
    avg_loss: float


MAX_STAGNANT_EPISODES = DMemDef.MAX_STAGNANT_EPISODES
MAX_HARD_RESET_EPISODES = DMemDef.MAX_HARD_RESET_EPISODES


class TrainMgr:
    """
    Training orchestrator.

    Owns:
      - policy (possibly epsilon-greedy wrapper)
      - trainer (torch + optimizer)
      - replay memory

    Uses:
      - SnakeMgr as the environment/session API

    Does NOT:
      - talk to MQ
      - render
    """

    def __init__(
        self,
        *,
        snake_mgr: SnakeMgr,
        policy: EpsilonNicePolicy,
        trainer: LinearTrainer | RecurrentTrainer,
        replay: SimpleReplayMemory | ATHMemory,
        client_id: str = DModule.TRAIN_MGR,
        model: LinearModel | RNNModel | GRUModel,
        log_level: DHydraLog,
        pub_func,
    ) -> None:
        self.event = HydraEventMQ(
            client_id=DModule.TRAIN_MGR,
            pub_func=pub_func,
        )
        self.log = HydraLog(
            client_id=DModule.TRAIN_MGR,
            log_level=log_level,
            to_console=True,
        )
        self.snake_mgr = snake_mgr
        self.policy = policy
        self.trainer = trainer
        self.replay = replay
        self.client_id = client_id
        self.model = model

        self._stag_alert_status = EV_TYPE.CLEARED
        self._stag_ep_count = 0
        self._hard_reset_ep_count = 0
        self._hard_reset_count = 0
        self._cur_highscore = 0
        self._max_hard_reset_episodes = MAX_HARD_RESET_EPISODES

    # Called at the end of each episode
    async def handle_stagnation(self, final_score):
        self.replay.gearbox.incr_epoch()

        if final_score > self._cur_highscore:
            self._cur_highscore = final_score
            self._stag_ep_count = 0
            self._hard_reset_ep_count = 0
            await self.replay.gearbox.stagnation_cleared()
            if self._stag_alert_status != EV_TYPE.CLEARED:
                msg = "Stagnation alert cleared"
                self.log.debug(msg)
                await self.event.publish(
                    EventMsg(level=EV_STATUS.GOOD, message=msg)
                )
                await self.policy.disable_nice()
            self._stag_alert_status = EV_TYPE.CLEARED

        else:
            self._stag_ep_count += 1
            self._hard_reset_ep_count += 1

        if self._stag_ep_count >= MAX_STAGNANT_EPISODES:
            if self._stag_alert_status != EV_TYPE.SET:
                msg = f"Stagnation alert raised"
                self.log.debug(msg)
                await self.event.publish(
                    EventMsg(level=EV_STATUS.WARN, message=msg)
                )
                self._stag_ep_count = 0
                self.replay.gearbox.stagnation_warning()
                await self.policy.enable_nice()
            self._stag_alert_status = EV_TYPE.SET

        if self._hard_reset_ep_count >= self._max_hard_reset_episodes:
            self._hard_reset_count += 1
            if self._hard_reset_count > 1:
                self._max_hard_reset_episodes = (
                    MAX_HARD_RESET_EPISODES * self._hard_reset_count
                )
                self.log.debug(
                    f"Critical stagnation alert({self._hard_reset_count}): "
                    f"Exceeded {self._hard_reset_count} episodes without "
                    "improvement"
                )
            await self.event.publish(
                EventMsg(
                    level=EV_STATUS.BAD,
                    message=f"Critical stagnation event({self._hard_reset_count}) detected",
                )
            )
            self._hard_reset_ep_count = 0
            self._stag_ep_count = 0
            await self.replay.gearbox.hard_reset(self._hard_reset_count)
            if self._stag_alert_status != EV_TYPE.SET:
                await self.policy.enable_nice()
            self._stag_alert_status = EV_TYPE.SET



# ai_hydra/nnet/SimpleReplayMemory.py
#
#    AI Hydra
#    Author: Nadim-Daniel Ghaznavi
#    Copyright: (c) 2025-2026 Nadim-Daniel Ghaznavi
#    GitHub: https://github.com/NadimGhaznavi/ai_hydra
#    Website: https://ai-hydra.readthedocs.io/en/latest
#    License: GPL 3.0

### ----- Version II -----

from __future__ import annotations

from collections import deque
from random import Random
from typing import Deque, TypeVar

from ai_hydra.constants.DReplayMemory import DMemory
from ai_hydra.constants.DHydra import DHydraLog, DModule

from ai_hydra.nnet.Transition import Transition
from ai_hydra.utils.HydraLog import HydraLog
from ai_hydra.zmq.HydraEventMQ import EventMsg, HydraEventMQ


MAX_MEM_SIZE = DMemory.MAX_MEM_SIZE  # Max Linear transitions
MIN_FRAMES = DMemory.MIN_FRAMES  # Min transitions before returning samples

T = TypeVar("T")


class SimpleReplayMemory:
    def __init__(
        self,
        rng: Random,
        log_level: DHydraLog,
        pub_func,
    ):
        self.log = HydraLog(
            client_id=DModule.SIMPLE_REPLAY_MEMORY,
            log_level=log_level,
            to_console=True,
        )
        self._rng = rng
        self.event = HydraEventMQ(
            client_id=DModule.SIMPLE_REPLAY_MEMORY, pub_func=pub_func
        )

        # Linear model settings
        self._memory: Deque[Transition] = deque(maxlen=DMemory.MAX_MEM_SIZE)
        self._memory_not_full = True
        self._memory_cold = True
        self._has_logged_startup = False

    async def _log_startup(self):
        self.log.info("Initialized")

        msg = f"Setting maximum number of stored transitions to {MAX_MEM_SIZE}"
        self.log.info(msg)
        await self.event.publish(EventMsg(level=DHydraLog.INFO, message=msg))

        msg = f"Setting minimum number of transitions to {MIN_FRAMES}"
        self.log.info(msg)
        await self.event.publish(EventMsg(level=DHydraLog.INFO, message=msg))

    async def append(self, t: Transition) -> None:
        """Add a transition into memory"""

        if not self._has_logged_startup:
            await self._log_startup()
            self._has_logged_startup = True

        self._memory.append(t)
        if len(self._memory) >= MAX_MEM_SIZE and self._memory_not_full:
            msg = f"Memory has been filled to capacity: {MAX_MEM_SIZE} transitions"
            self.log.info(msg)
            await self.event.publish(
                EventMsg(level=DHydraLog.INFO, message=msg)
            )
            self._memory_not_full = False

    async def sample_transitions(
        self, batch_size: int
    ) -> list[Transition] | None:
        """Sample random transitions for a linear model."""

        if len(self._memory) < max(MIN_FRAMES, batch_size):
            return None

        if self._memory_cold:
            msg = "Memory has warmed up"
            self.log.debug(msg)
            await self.event.publish(
                EventMsg(level=DHydraLog.INFO, message=msg)
            )
            self._memory_cold = False

        return self._rng.sample(list(self._memory), batch_size)



# ai_hydra/nnet/ATH/ATHDataMgr.py
#
#    AI Hydra
#    Author: Nadim-Daniel Ghaznavi
#    Copyright: (c) 2025-2026 Nadim-Daniel Ghaznavi
#    GitHub: https://github.com/NadimGhaznavi/ai_hydra
#    Website: https://ai-hydra.readthedocs.io/en/latest
#    License: GPL 3.0

from random import Random

from ai_hydra.constants.DReplayMemory import DMemDef, DMemField
from ai_hydra.constants.DHydra import DHydraLog, DModule
from ai_hydra.constants.DEvent import EV_TYPE, EV_STATUS

from ai_hydra.utils.HydraLog import HydraLog
from ai_hydra.zmq.HydraEventMQ import HydraEventMQ, EventMsg
from ai_hydra.nnet.Transition import Transition
from ai_hydra.nnet.ATH.ATHDataStore import ATHDataStore
from ai_hydra.nnet.ATH.ATHCommon import (
    GearMeta,
    get_gear_data,
    get_valid_gears,
)


MAX_BUCKETS = DMemDef.MAX_BUCKETS
MAX_FRAMES = DMemDef.MAX_FRAMES


class ATHDataMgr:

    def __init__(
        self,
        log_level: DHydraLog,
        rng: Random,
        pub_func,
        data_store: ATHDataStore,
    ):
        self._rng = rng
        self.store = data_store
        # Local logging
        self.log = HydraLog(
            client_id=DModule.ATH_DATA_MGR,
            log_level=log_level,
            to_console=True,
        )
        # Remote logging: ZeroMQ "Events" PUB/SUB topic
        self.event = HydraEventMQ(
            client_id=DModule.ATH_DATA_MGR, pub_func=pub_func
        )

        self._cur_gear = None
        self._samples_served = 0
        self._has_logged_startup = False
        self._has_logged_pruning = False

    async def append(
        self, t: Transition, final_score: int | None = None
    ) -> None:

        if not self._has_logged_startup:
            await self._log_startup()
            self._has_logged_startup = True

        self.store.append(t)

        if t.done:
            if final_score is None:
                raise ValueError(
                    "final_score must be provided when appending a terminal transition"
                )

            await self._finalize_game()

    def _build_episode_gear_meta(
        self,
        ep_size: int,
    ) -> dict[int, GearMeta]:
        """
        Build metadata for all gears for one episode.
        """
        gear_meta: dict[int, GearMeta] = {}

        for gear in get_valid_gears():
            seq_length, _ = get_gear_data(gear)
            gear_meta[gear] = self._build_gear_meta_for_size(
                ep_size=ep_size,
                seq_length=seq_length,
            )

        return gear_meta

    def _build_gear_meta_for_size(
        self,
        ep_size: int,
        seq_length: int,
    ) -> GearMeta:
        """
        Build gear-specific chunk metadata for an episode size.

        Chunks are aligned from the end, matching the old ReplayMemory behavior.
        """
        num_chunks = ep_size // seq_length
        rem = ep_size % seq_length

        ranges: list[tuple[int, int]] = []

        for chunk_idx in range(num_chunks):
            start = rem + (chunk_idx * seq_length)
            end = start + seq_length
            ranges.append((start, end))

        return GearMeta(
            seq_length=seq_length,
            num_chunks=num_chunks,
            ranges=tuple(ranges),
        )

    async def _finalize_game(self) -> None:
        ep_size = self.store.get_cur_ep_size()
        gear_meta = self._build_episode_gear_meta(ep_size)
        self.store.finalize_game(gear_meta)

        await self._prune_if_needed()
        self._rebuild_bucket_index()

    async def _log_startup(self):
        self.log.info(
            f"Initialized. Setting max_frames to {MAX_FRAMES}, max_buckets to {MAX_BUCKETS}"
        )

    async def _prune_if_needed(self) -> None:

        while self.store.get_stored_frame_count() > MAX_FRAMES:
            self.store.pop_first_game()

            if not self._has_logged_pruning:
                msg = f"Memory is full ({MAX_FRAMES}), pruning initiated"
                self.log.info(msg)

                await self.event.publish(
                    EventMsg(level=EV_STATUS.INFO, message=msg)
                )
                self._has_logged_pruning = True

    def _rebuild_bucket_index(self):
        self.store.reset_bucket_indexes()
        self.store.update_episodes_by_bucket()
        self.store.update_warmed_buckets_list()

    async def sample_chunks(self) -> list[list[Transition]] | None:
        """
        Sample chunks across warmed buckets for the current gear.
        """
        warmed_buckets = self.store.get_warmed_buckets()

        if not warmed_buckets:
            return None

        total_chunks = 0
        for bucket_idx in warmed_buckets:
            episodes = self.store.get_eps_by_bucket_idx(bucket_idx=bucket_idx)
            total_chunks += len(episodes)

        _, batch_size = get_gear_data(self._cur_gear)

        if total_chunks < batch_size:
            return None

        samples: list[list[Transition]] = []

        base = batch_size // len(warmed_buckets)
        remainder = batch_size % len(warmed_buckets)

        requested: dict[int, int] = {idx: base for idx in warmed_buckets}

        warmed_desc = sorted(warmed_buckets, reverse=True)
        for i in range(remainder):
            requested[warmed_desc[i % len(warmed_desc)]] += 1

        for bucket_idx in warmed_buckets:
            eligible = self.store.get_eps_by_bucket_idx(bucket_idx=bucket_idx)
            if not eligible:
                continue

            meta_samples = requested[bucket_idx]

            for _ in range(meta_samples):
                ep = self._rng.choice(eligible)
                meta = ep.gear_meta[self._cur_gear]
                start, end = meta.ranges[bucket_idx]
                samples.append(list(ep.frames[start:end]))

        self._samples_served += 1

        if self._samples_served % 10 == 0:
            bucket_counts = self.store.get_bucket_counts()
            await self.event.publish(
                EventMsg(
                    level=EV_STATUS.INFO,
                    ev_type=EV_TYPE.BUCKETS_STATUS,
                    payload={
                        EV_TYPE.BUCKET_COUNTS: bucket_counts,
                    },
                )
            )
        return samples

    def set_gear(self, gear: int):
        if self._cur_gear == gear:
            return

        self._cur_gear = gear
        self.store.set_gear(gear)
        self._rebuild_bucket_index()



# ai_hydra/client/PongBoard.py
#
#    AI Hydra
#    Author: Nadim-Daniel Ghaznavi
#    Copyright: (c) 2025-2026 Nadim-Daniel Ghaznavi
#    GitHub: https://github.com/NadimGhaznavi/ai_hydra
#    Website: https://ai-hydra.readthedocs.io/en/latest
#    License: GPL 3.0

from textual.scroll_view import ScrollView
from textual.geometry import Offset, Region, Size
from textual.strip import Strip
from textual.reactive import var

from rich.segment import Segment
from rich.style import Style

from ai_hydra.constants.DGame import DGameField

emptyA = "#111111"
emptyB = "#000000"
food = "#940101"
snake = "#025b02"
snake_head = "#16e116"


class PongBoard(ScrollView):
    COMPONENT_CLASSES = {
        "clientgameboard--emptyA-square",
        "clientgameboard--emptyB-square",
        "clientgameboard--food-square",
        "clientgameboard--snake-square",
        "clientgameboard--snake-head-square",
    }

    DEFAULT_CSS = """
    ClientGameBoard > .clientgameboard--emptyA-square {
        background: #111111;
    }
    ClientGameBoard > .clientgameboard--emptyB-square {
        background: #000000;
    }
    ClientGameBoard > .clientgameboard--food-square {
        background: #940101;
    }
    ClientGameBoard > .clientgameboard--snake-square {
        background: #025b02;
    }
    ClientGameBoard > .clientgameboard--snake-head-square {
        background: #0ca30c;
    }
    """

    # Reactive state (screen coords: Offset(x,y))
    food = var(Offset(9, 9))
    snake_head = var(Offset())
    snake_body = var([])
    direction = var(Offset(1, 0))

    def __init__(self, board_size: int, id=None) -> None:
        super().__init__(id=id)
        self._board_size = int(board_size)
        self.virtual_size = Size(board_size * 2, board_size)

    def board_size(self) -> int:
        return self._board_size

    # Updates
    # --------
    def apply_board_dict(self, board: dict) -> None:
        grid = board[DGameField.GRID_SIZE]
        self.grid_w = int(grid[DGameField.W])
        self.grid_h = int(grid[DGameField.H])

        head = board[DGameField.SNAKE_HEAD]
        self.snake_head = Offset(
            int(head[DGameField.X]), int(head[DGameField.Y])
        )

        self.snake_body = [
            Offset(int(seg[DGameField.X]), int(seg[DGameField.Y]))
            for seg in board[DGameField.SNAKE_BODY]
        ]

        food = board[DGameField.FOOD_POSITION]
        self.food = Offset(int(food[DGameField.X]), int(food[DGameField.Y]))

        d = board[DGameField.DIRECTION]
        self.direction = Offset(int(d[DGameField.DX]), int(d[DGameField.DY]))

        self.refresh()

    # Rendering
    # ----------
    def get_square_region(self, square_offset: Offset) -> Region:
        """Get region relative to widget from square coordinate."""
        x, y = square_offset
        region = Region(x * 2, y, 2, 1)
        # Move the region into the widgets frame of reference
        return region.translate(-self.scroll_offset)

    def render_line(self, y: int) -> Strip:
        scroll_x, scroll_y = self.scroll_offset
        y += scroll_y
        row_index = y

        emptyA = self.get_component_rich_style(
            "clientgameboard--emptyA-square"
        )
        emptyB = self.get_component_rich_style(
            "clientgameboard--emptyB-square"
        )
        food = self.get_component_rich_style("clientgameboard--food-square")
        snake = self.get_component_rich_style("clientgameboard--snake-square")
        snake_head = self.get_component_rich_style(
            "clientgameboard--snake-head-square"
        )

        if row_index >= self._board_size:
            return Strip.blank(self.size.width)

        is_odd = row_index % 2

        def get_square_style(column: int, row: int) -> Style:
            pos = Offset(column, row)
            if self.food == pos:
                return food
            if self.snake_head == pos:
                return snake_head
            if pos in self.snake_body:
                return snake
            return emptyA if (column + is_odd) % 2 else emptyB

        segments = [
            Segment(" " * 2, get_square_style(column, row_index))
            for column in range(self._board_size)
        ]
        strip = Strip(segments, self._board_size * 2)
        # Crop the strip so that is covers the visible area
        return strip.crop(scroll_x, scroll_x + self.size.width)

    # Reactive watchers
    # ------------------
    def watch_food(self, previous_food, food) -> None:
        """Called when the food square changes."""
        # Refresh the previous food square
        self.refresh(self.get_square_region(previous_food))
        # Refresh the new food square
        self.refresh(self.get_square_region(food))

    def watch_snake_head(
        self, previous_snake_head: Offset, snake_head: Offset
    ) -> None:
        """Called when the snake head square changes."""
        self.refresh(self.get_square_region(previous_snake_head))
        self.refresh(self.get_square_region(snake_head))

    def watch_snake_body(
        self, previous_snake_body: list, snake_body: list
    ) -> None:
        """Called when the snake body changes."""
        for segment in previous_snake_body:
            self.refresh(self.get_square_region(segment))

        for segment in snake_body:
            self.refresh(self.get_square_region(segment))



# ai_hydra/client/EventLog.py
#
#    AI Hydra
#    Author: Nadim-Daniel Ghaznavi
#    Copyright: (c) 2025-2026 Nadim-Daniel Ghaznavi
#    GitHub: https://github.com/NadimGhaznavi/ai_hydra
#    Website: https://ai-hydra.readthedocs.io/en/latest
#    License: GPL 3.0

from textual.containers import Vertical
from textual.widgets import TabbedContent, Label, Log
from textual.app import ComposeResult, Widget

from ai_hydra.constants.DHydra import DHydraLog
from ai_hydra.constants.DHydraTui import DField, DLabel
from ai_hydra.constants.DEvent import EVENT_MAP, EV_STATUS

from ai_hydra.utils.HydraMetrics import HydraMetrics

LEVEL_MAP = {
    DHydraLog.INFO: EV_STATUS.INFO,
    DHydraLog.WARNING: EV_STATUS.WARN,
    DHydraLog.ERROR: EV_STATUS.BAD,
}


class EventLog(Widget):

    def __init__(self, metrics: HydraMetrics, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self.metrics = metrics

    def compose(self) -> ComposeResult:
        yield Vertical(
            Label(
                f"[b #3e99af]{'Epoch':>7s}{DLabel.TIME:>13s}    {'Source':<17s}{'  ':<2}{DLabel.EVENT}[/]"
            ),
            Log(highlight=True, auto_scroll=True, id=DField.EVENT_LOG_LOG),
        )

    def clear(self):
        self.query_one(f"#{DField.EVENT_LOG_LOG}", Log).clear()

    def add_event(
        self, ev_type: str, event: str, status: str = None, epoch: str = None
    ):
        if epoch:
            epoch = str(epoch)
        else:
            epoch = "-"

        if not status:
            status = EV_STATUS.INFO

        elap_time = self.metrics.get_elapsed_time()
        src_icon, source = EVENT_MAP[ev_type]

        msg = f"{epoch:>7s}{elap_time:>13s}  {src_icon:>3s} {source:<16s}{status:<2}{event}"

        self.query_one(f"#{DField.EVENT_LOG_LOG}", Log).write_line(msg)

        self.metrics.add_event_log_msg(msg)



# ai_hydra/server/HydraMgr.py
#
#    AI Hydra
#    Author: Nadim-Daniel Ghaznavi
#    Copyright: (c) 2025-2026 Nadim-Daniel Ghaznavi
#    GitHub: https://github.com/NadimGhaznavi/ai_hydra
#    Website: https://ai-hydra.readthedocs.io/en/latest
#    License: GPL 3.0

from __future__ import annotations
import asyncio
import argparse
import traceback
from datetime import datetime

from ai_hydra.constants.DGame import DGameMethod, DGameField
from ai_hydra.constants.DHydra import (
    DHydraLog,
    DModule,
    DHydraLogDef,
    DHydraRouterDef,
    DHydraServerDef,
    DHydra,
)
from ai_hydra.constants.DNNet import DNetField
from ai_hydra.constants.DHydra import DMethod
from ai_hydra.constants.DHydraTui import DField

from ai_hydra.server.HydraServer import HydraServer
from ai_hydra.server.SnakeMgr import SnakeMgr
from ai_hydra.zmq.HydraMsg import HydraMsg
from ai_hydra.nnet.Transition import Transition
from ai_hydra.nnet.HydraRng import HydraRng
from ai_hydra.utils.SimCfg import SimCfg


class HydraMgr(HydraServer):
    def __init__(
        self,
        address: str = "*",
        port: int = DHydraServerDef.PORT,
        router_address: str = DHydraRouterDef.HOSTNAME,
        router_port: int = DHydraRouterDef.PORT,
        identity: str = DModule.HYDRA_MGR,
        log_level: DHydraLog = DHydraLogDef.DEFAULT_LOG_LEVEL,
        debug: bool = False,
    ) -> None:
        super().__init__(
            address=address,
            port=port,
            router_address=router_address,
            router_port=router_port,
            identity=identity,
            log_level=log_level,
        )

        self.debug = bool(debug)
        self.cfg = SimCfg()
        self.hydra_rng: HydraRng | None = None
        self.snake: SnakeMgr | None = None
        self._train_mgr = None
        self._train_mgr_model_type: str | None = None
        self._methods.update(
            {
                DMethod.HANDSHAKE: self.handshake,
                DGameMethod.PAUSE_RUN: self._pause_run,
                DGameMethod.RESET_RUN: self._reset_run,
                DGameMethod.RESUME_RUN: self._resume_run,
                DGameMethod.START_RUN: self.start_run,
                DGameMethod.STOP_RUN: self.stop_run,
                DGameMethod.UPDATE_CONFIG: self.update_config,
            }
        )
        if self.debug:
            self._methods[DGameMethod.GAME_STEP] = self.game_step

        self._runs: dict[str, asyncio.Task[None]] = {}
        self._client_id = None
        self._sim_running = False
        self._sim_paused = False

        self._run_loop_task: asyncio.Task[None] | None = None
        self._run_loop_pause_event = asyncio.Event()
        self._run_loop_stop_event = asyncio.Event()

    def _ensure_train_mgr(self):
        """
        Lazily construct TrainMgr and NN components.

        This keeps HydraMgr import graph clean for headless server use,
        and avoids circular imports until training is actually invoked.
        """

        # Late imports to avoid circularities
        import torch

        from ai_hydra.nnet.TrainMgr import TrainMgr

        from ai_hydra.nnet.ATH.ATHMemory import ATHMemory
        from ai_hydra.nnet.SimpleReplayMemory import SimpleReplayMemory

        from ai_hydra.nnet.LinearTrainer import LinearTrainer
        from ai_hydra.nnet.RecurrentTrainer import RecurrentTrainer

        from ai_hydra.nnet.models.GRUModel import GRUModel
        from ai_hydra.nnet.models.LinearModel import LinearModel
        from ai_hydra.nnet.models.RNNModel import RNNModel

        from ai_hydra.nnet.Policy.LinearPolicy import LinearPolicy
        from ai_hydra.nnet.Policy.RecurrentPolicy import RecurrentPolicy
        from ai_hydra.nnet.Policy.EpsilonPolicy import EpsilonPolicy
        from ai_hydra.nnet.Policy.EpsilonNicePolicy import EpsilonNicePolicy

        from ai_hydra.nnet.EpsilonAlgo import EpsilonAlgo
        from ai_hydra.nnet.EpsilonNiceAlgo import EpsilonNiceAlgo

        from ai_hydra.constants.DNNet import DNetField, DLinear, DRNN, DGRU

        model_type = self.cfg.get(DNetField.MODEL_TYPE)
        master_seed = self.cfg.get(DNetField.RANDOM_SEED)

        self.hydra_rng = HydraRng(
            log_level=self.log_level,
            pub_func=self.mq.publish_events,
            master_seed=master_seed,
        )
        master_seed = self.hydra_rng.get_seed()

        # Hydra-style RNG streams
        _, policy_rng = self.hydra_rng.new_rng()
        _, replay_rng = self.hydra_rng.new_rng()

        device = torch.device(
            DField.CPU
        )  # keep simple; GPU can be passed later

        if model_type == DField.LINEAR:
            self.log.debug("Using Linear Model")
            epsilon_nice_p_val = DLinear.NICE_P_VALUE
            replay = SimpleReplayMemory(
                rng=replay_rng,
                log_level=self.log_level,
                pub_func=self.mq.publish_events,
            )
            model = LinearModel(log_level=self.log_level, seed=master_seed)
            model.set_params(
                hidden_size=self.cfg.get(DNetField.HIDDEN_SIZE),
                dropout_p=self.cfg.get(DNetField.DROPOUT_P),
            )
            nnet_policy = LinearPolicy(model=model, device=device)
            trainer = LinearTrainer(
                model=model,
                replay=replay,
                lr=self.cfg.get(DNetField.LEARNING_RATE),
                device=device,
                log_level=self.log_level,
            )
            trainer.set_params(
                gamma=self.cfg.get(DNetField.GAMMA),
                batch_size=self.cfg.get(DNetField.BATCH_SIZE),
            )

        elif model_type == DField.RNN:
            self.log.debug("Using RNN Model")
            epsilon_nice_p_val = DRNN.NICE_P_VALUE
            epsilon_nice_steps = DRNN.NICE_STEPS
            model = RNNModel(log_level=self.log_level, seed=master_seed)
            model.set_params(
                hidden_size=self.cfg.get(DNetField.HIDDEN_SIZE),
                dropout_p=self.cfg.get(DNetField.DROPOUT_P),
                rnn_layers=self.cfg.get(DNetField.LAYERS),
            )
            nnet_policy = RecurrentPolicy(model=model, device=device)
            replay = ATHMemory(
                rng=replay_rng,
                log_level=self.log_level,
                pub_func=self.mq.publish_events,
            )
            trainer = RecurrentTrainer(
                model=model,
                replay=replay,
                lr=self.cfg.get(DNetField.LEARNING_RATE),
                device=device,
                log_level=self.log_level,
            )
            trainer.set_params(
                tau=self.cfg.get(DNetField.TAU),
                gamma=self.cfg.get(DNetField.GAMMA),
            )

        elif model_type == DField.GRU:
            self.log.debug("Using GRU Model")
            epsilon_nice_p_val = DGRU.NICE_P_VALUE
            epsilon_nice_steps = DGRU.NICE_STEPS
            model = GRUModel(log_level=self.log_level, seed=master_seed)
            model.set_params(
                hidden_size=self.cfg.get(DNetField.HIDDEN_SIZE),
                dropout_p=self.cfg.get(DNetField.DROPOUT_P),
                gru_layers=self.cfg.get(DNetField.LAYERS),
            )
            nnet_policy = RecurrentPolicy(model=model, device=device)
            replay = ATHMemory(
                rng=replay_rng,
                log_level=self.log_level,
                pub_func=self.mq.publish_events,
            )
            trainer = RecurrentTrainer(
                model=model,
                replay=replay,
                lr=self.cfg.get(DNetField.LEARNING_RATE),
                device=device,
                log_level=self.log_level,
            )
            trainer.set_params(
                tau=self.cfg.get(DNetField.TAU),
                gamma=self.cfg.get(DNetField.GAMMA),
            )

        else:
            raise ValueError(f"Unsupported model type: {model_type}")

        self.log.debug(f"Model details:")
        self.log.debug(str(model))

        # Policy stack
        epsilon_algo = EpsilonAlgo(
            rng=policy_rng,
            log_level=self.log_level,
            pub_func=self.mq.publish_events,
        )
        epsilon_algo.initial_epsilon(self.cfg.get(DNetField.INITIAL_EPSILON))
        epsilon_algo.min_epsilon(self.cfg.get(DNetField.MIN_EPSILON))
        epsilon_algo.decay_rate(self.cfg.get(DNetField.EPSILON_DECAY))

        epsilon_nice = EpsilonNiceAlgo(
            log_level=self.log_level,
            pub_func=self.mq.publish_events,
            rng=policy_rng,
        )
        epsilon_policy = EpsilonPolicy(
            base_policy=nnet_policy, epsilon=epsilon_algo
        )

        behaviour_policy = EpsilonNicePolicy(
            base_policy=epsilon_policy,
            epsilon_n=epsilon_nice,
            p_value=epsilon_nice_p_val,
            rng=policy_rng,
            steps=epsilon_nice_steps,
            log_level=self.log_level,
            pub_func=self.mq.publish_events,
        )

        self._train_mgr = TrainMgr(
            snake_mgr=self.snake,
            policy=behaviour_policy,
            trainer=trainer,
            replay=replay,
            client_id=DModule.TRAIN_MGR,
            model=model,
            log_level=self.log_level,
            pub_func=self.mq.publish_events,
        )
        self._train_mgr_model_type = model_type
        return self._train_mgr

    async def handshake(self, msg: HydraMsg) -> None:
        """
        When a HydraClient starts, it sends a "handshake".
        """
        try:
            payload = self.cfg.to_dict()
            if self._sim_running:
                payload[DGameField.SIM_RUNNING] = True
            else:
                payload[DGameField.SIM_RUNNING] = False

            if self._sim_paused:
                payload[DGameField.SIM_PAUSED] = True
            else:
                payload[DGameField.SIM_PAUSED] = False

            reply = HydraMsg(
                sender=self.identity,
                target=msg.sender,
                method=DMethod.HANDSHAKE_REPLY,
                payload=payload,
            )

            if self.mq is not None:
                await self.mq.send(reply)
        except Exception as e:
            self.log.critical(f"ERROR: {e}")
            self.log.critical(f"TRACEBACK: {traceback.format_exc()}")

    async def _pause_run(self, msg: HydraMsg) -> None:
        """
        Pause a running simulation.
        """
        self._run_loop_pause_event.set()
        self._sim_paused = True
        reply = HydraMsg(
            sender=self.identity,
            target=msg.sender,
            method=DGameMethod.PAUSE_RUN_REPLY,
        )
        if self.mq is not None:
            await self.mq.send(reply)

    async def _resume_run(self, msg: HydraMsg) -> None:
        """
        Resume a paused simulation.
        """
        self._sim_paused = False
        self._run_loop_pause_event.clear()
        reply = HydraMsg(
            sender=self.identity,
            target=msg.sender,
            method=DGameMethod.RESUME_RUN_REPLY,
        )
        if self.mq is not None:
            await self.mq.send(reply)

    async def _reset_run(self, msg: HydraMsg) -> None:
        """
        Stop the running simulation.
        """
        self._run_loop_stop_event.set()

        reply = HydraMsg(
            sender=self.identity,
            target=msg.sender,
            method=DGameMethod.RESET_RUN_REPLY,
        )
        if self.mq is not None:
            await self.mq.send(reply)

        # Wait for the simulation to end
        while self._sim_running:
            await asyncio.sleep(0.1)

        # Reset the model's weights
        # self._train_mgr.model.reset_weights()
        # Reset the optimizer
        # self._train_mgr.trainer.reset()

        self._sim_paused = False
        self._runs = {}
        self._run_loop_pause_event.clear()
        self._run_loop_stop_event.clear()
        self.log.info("Stopping the simulation and resetting the environment")

    async def _run_loop(self, client_id: str) -> None:
        """
        Runs as fast as possible.
        """
        self.log.debug("Starting simulation run")
        self.log.debug(
            f"Using random seed: {self.cfg.get(DNetField.RANDOM_SEED)}"
        )
        try:
            model_type = self.cfg.get(DNetField.MODEL_TYPE)

            train_mgr = self._ensure_train_mgr()
            snake = self.snake = SnakeMgr(
                cfg=self.cfg,
                log_level=self.log_level,
                hydra_rng=self.hydra_rng,
            )
            mq = self.mq
            train_mgr.policy.reset_episode()

            sess = snake.get_session(client_id)
            snake.start_time(datetime.now())

            count = 0
            train_every = 4
            grad_steps = 1
            batch_size = 64

            try:
                while not self._run_loop_stop_event.is_set():

                    while self._run_loop_pause_event.is_set():
                        await asyncio.sleep(0.1)

                    # Auto-reset when done
                    if sess.done:
                        snake.reset_session(client_id)
                        sess = snake.get_session(client_id)
                        train_mgr.policy.reset_episode()

                    # Decide if we will publish per_step *this tick*
                    want_step = self.cfg.get(DNetField.PER_STEP)

                    # Choose action (server-side)
                    old_state = sess.board.get_state()

                    # EpsilonNice needs the board
                    action = train_mgr.policy.select_action(
                        old_state, sess.board
                    )

                    # Advance sim
                    state_dict, scores_payload, step_payload, ep_payload = (
                        snake.step(
                            client_id=client_id,
                            action=action,
                        )
                    )

                    done = state_dict[DGameField.DONE]

                    # Episode-end bookkeeping
                    if done:
                        sess.epoch += 1
                        count += 1

                        if count % 100 == 0:
                            self.log.info(f"Epoch: {count}")

                        # Epsilon
                        await train_mgr.policy.played_game()
                        ep_payload[DNetField.CUR_EPSILON] = (
                            train_mgr.policy.cur_epsilon()
                        )

                        if (
                            model_type == DField.RNN
                            or model_type == DField.GRU
                        ):
                            # train_mgr.replay.append(t=t, final_score=sess.score)
                            await train_mgr.trainer.train_long_memory()
                            # Anti-Stagnation strategy
                            if DNetField.FINAL_SCORE in scores_payload:
                                await train_mgr.handle_stagnation(
                                    scores_payload[DNetField.FINAL_SCORE]
                                )

                        # Loss, if available, is loaded into the telemetry here
                        loss = train_mgr.trainer.get_avg_loss()
                        if loss is not None:
                            ep_payload[DNetField.LOSS] = loss

                    reward = state_dict[DGameField.REWARD]
                    new_state = state_dict[DNetField.NEXT_STATE]

                    # Build/store transition (unchanged for now)
                    t = Transition(
                        old_state=tuple(old_state),
                        action=int(action),
                        reward=float(reward),
                        new_state=tuple(new_state),
                        done=bool(done),
                    )

                    if model_type == DField.RNN or model_type == DField.GRU:
                        await train_mgr.replay.append(
                            t=t, final_score=sess.score
                        )
                    else:
                        await train_mgr.replay.append(t=t)

                    if model_type == DField.LINEAR:
                        if sess.step_n % train_every == 0:
                            for _ in range(grad_steps):
                                await train_mgr.trainer.train_long_memory()

                    # Publish
                    if mq is not None:
                        # Publish scores at every step
                        await mq.publish_scores(scores_payload)

                        # Publish per epsisode if the episode is "done"
                        if done:
                            await mq.publish_per_episode(ep_payload)
                        # publish per_step only if enabled and payload exists
                        if want_step and step_payload:
                            await mq.publish_per_step(step_payload)

                    # Yield to event loop so MQ IO stays healthy
                    delay = (
                        0.0
                        if not want_step
                        else self.cfg.get(DNetField.MOVE_DELAY)
                    )
                    await asyncio.sleep(delay)

                self._sim_running = False

            except asyncio.CancelledError:
                raise
            except Exception as e:
                self.log.critical(f"ERROR: {e}")
                self.log.critical(f"STACKTRACE: {traceback.format_exc()}")

        except asyncio.CancelledError:
            return
        except Exception as e:
            self.log.critical(f"ERROR: {e}")
            self.log.critical(f"STACKTRACE: {traceback.format_exc()}")

    async def start_run(self, msg: HydraMsg) -> None:
        self._sim_running = True
        client_id = msg.sender
        self._client_id = client_id
        self.log.debug(f"Received START_RUN {msg.payload}")

        try:

            # Get runtime settings
            self.cfg = SimCfg.from_dict(msg.payload)

            # already running?
            if client_id in self._runs and not self._runs[client_id].done():
                payload = {
                    DGameField.OK: False,
                    DGameField.INFO: {"status": "already_running"},
                }
            else:
                self._run_loop_task = asyncio.create_task(
                    self._run_loop(client_id)
                )
                self._runs[client_id] = self._run_loop_task
                payload = {
                    DGameField.OK: True,
                    DGameField.INFO: {"status": "started"},
                }

            reply = HydraMsg(
                sender=self.identity,
                target=client_id,
                method=DGameMethod.START_RUN,
                payload=payload,
            )
            if self.mq is not None:
                await self.mq.send(reply)

        except Exception as e:
            self.log.critical(f"ERROR: {e}")
            self.log.critical(f"STACKTRACE: {traceback.format_exc()}")

    async def stop_run(self, msg: HydraMsg) -> None:
        client_id = msg.sender

        task = self._runs.get(client_id)
        if task and not task.done():
            task.cancel()
            payload = {
                DGameField.OK: True,
                DGameField.INFO: {"status": "stopping"},
            }
        else:
            payload = {
                DGameField.OK: True,
                DGameField.INFO: {"status": "not_running"},
            }

        reply = HydraMsg(
            sender=self.identity,
            target=client_id,
            method=DGameMethod.STOP_RUN,
            payload=payload,
        )
        if self.mq is not None:
            await self.mq.send(reply)

    async def update_config(self, msg: HydraMsg) -> None:
        """
        Update settings while a simulation is running
        """
        self.log.debug(f"Received config update")
        # Currently only per_step (on/off) and move_delay are settable
        # from the TUI when a sim is running. Also, the TUI can only
        # send the update when the sim is running.
        self.cfg.set(
            key=DNetField.PER_STEP, value=msg.payload[DNetField.PER_STEP]
        )
        self.cfg.set(
            key=DNetField.MOVE_DELAY, value=msg.payload[DNetField.MOVE_DELAY]
        )


async def amain() -> None:
    p = argparse.ArgumentParser(description="AI Hydra Manager")
    p.add_argument("--address", default="*", help="Bind address")
    p.add_argument("--port", type=int, default=DHydraServerDef.PORT)
    p.add_argument("--router-address", default=DHydraRouterDef.HOSTNAME)
    p.add_argument("--router-port", type=int, default=DHydraRouterDef.PORT)
    p.add_argument("--identity", default=DModule.HYDRA_MGR)
    p.add_argument("--log-level", default=DHydraLogDef.DEFAULT_LOG_LEVEL)
    args = p.parse_args()

    server = HydraMgr(
        address=args.address,
        port=args.port,
        router_address=args.router_address,
        router_port=args.router_port,
        identity=args.identity,
        log_level=args.log_level,
    )

    await server.run()


def main() -> None:
    try:
        asyncio.run(amain())
    except BaseException:
        traceback.print_exc()
        raise


if __name__ == "__main__":
    main()
